(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Check_TP
imports Check_DPP 
begin

subsection {*Proving termination of TRSs*}

datatype ('f, 'l, 'v) trs_proof =
  DPTrans "(('f, 'l) lab, 'v) rules" "('f, 'l, 'v) dp_proof"
| RuleRemoval "('f, 'l) lab redtriple_impl" "('f, 'l, 'v) trsLL" "('f, 'l, 'v)trs_proof"
| StringReversal "('f, 'l, 'v) trs_proof"
| Bounds "(('f, 'l) lab, 'v) bounds_info"
| Uncurry "(('f, 'l) lab, 'v) uncurry_info" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trs_proof"
| Semlab
    "(('f, 'l) lab, 'v) sl_variant" "(('f, 'l) lab, 'v) term list"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trs_proof"
| RisEmpty
| Fcc "(('f, 'l) lab, 'v) ctxt list" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trs_proof"
| Split "('f, 'l, 'v) trsLL" "('f, 'l, 'v)trs_proof" "('f, 'l, 'v)trs_proof"
| SwitchInnermost "(('f,'l)lab, 'v)crit_pair_joins option" "('f, 'l, 'v)trs_proof" 
| AssumeSN "('f, 'l, 'v) qreltrsLL"

fun
  check_tp_subsumes ::
    "('tp, ('f::show, 'l::show) lab, 'v::show) tp_ops \<Rightarrow> ('f, 'l, 'v) qreltrsLL \<Rightarrow>
    'tp \<Rightarrow> shows check"
where
  "check_tp_subsumes I (q', r', rw') tp = (do {
    let q   = tp_ops.Q I tp;
    let r   = tp_ops.R I tp;
    let rw  = tp_ops.Rw I tp;
    let rb' = r' @ rw';
    check_NF_terms_subset q q' <+? (\<lambda>r. shows ''problem with innermost strategy due to term '' +@+ shows_term r);
    check_subseteq r r'   <+? (\<lambda>r. toomuch ''rule'' (shows_rule r));
    check_subseteq rw rb' <+? (\<lambda>r. toomuch ''relative rule'' (shows_rule r));
    return ()
  }) <+? (\<lambda>s. ''termination of the problem'' +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl
           +@+ ''may not be concluded from assuming termination of the problem'' +#+ shows_nl
           +@+ shows_tp I (tp_ops.mk I q' r' rw') +@+ shows_nl
           +@+ s +@+ shows_nl)"

lemma check_tp_subsumesE[elim]:
  assumes 1: "tp_spec I"
    and 2: "isOK (check_tp_subsumes I (q, r, rw) t)"
    and 3: "NF_terms (set (tp_ops.Q I t)) \<subseteq> NF_terms (set q)
    \<Longrightarrow> set (tp_ops.R I t) \<subseteq> set r
    \<Longrightarrow> set (tp_ops.Rw I t) \<subseteq> set r \<union> set rw
    \<Longrightarrow> P"
  shows "P"
  by (rule 3, insert 1 2) (simp add: Let_def | blast)+

fun SN_qrelL :: "('f, 'l, 'v) qreltrsLL \<Rightarrow> bool" where
  "SN_qrelL (q, r, rw) = SN_qrel (set q, set r, set rw)"

lemma check_tp_subsumes_sound:
  assumes I: "tp_spec I"
    and ok: "isOK (check_tp_subsumes I tp' tp)"
    and fin: "SN_qrelL tp'"
  shows "SN_qrel (tp_ops.qreltrs I tp)"
proof -
  obtain q r rw where tp': "tp' = (q, r, rw)" by (cases tp') force+
  let ?Q   = "set (tp_ops.Q I tp)"
  let ?R   = "set (tp_ops.R I tp)"
  let ?Rw  = "set (tp_ops.Rw I tp)"
  let ?Q'  = "set q"
  let ?R'  = "set r"
  let ?Rw' = "set rw"
  from I and ok
    have q: "NF_terms ?Q \<subseteq> NF_terms ?Q'" and "?R \<subseteq> ?R'" and "?Rw \<subseteq> ?R' \<union> ?Rw'"
    unfolding tp' by blast+
  note 1 = this
  show ?thesis unfolding tp_spec.qreltrs_sound[OF I]
    by (rule  SN_qrel_mono[OF q _ _ fin[unfolded tp', simplified]]) 
       (insert 1, blast+)
qed

fun
  check_trs_proof ::
    "bool \<Rightarrow> shows \<Rightarrow>
     ('tp, ('f::{show,linorder}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> 'tp \<Rightarrow>
     ('f, label_type, string) trs_proof \<Rightarrow>
     shows check" 
where
  "check_trs_proof a i I J tp RisEmpty = debug (i []) ''R is empty''
     (if tp_ops.R I tp = []
          then return () 
          else error (i +@+ '': R is not empty in the following termination-problem '' +#+
         shows_nl +@+ shows_tp I tp))"
|  "check_trs_proof a i I J tp (RuleRemoval redp rR prf) = debug (i []) ''Rule Removal'' (do {
         let R = tp_ops.rules I tp;
         let Rr = list_diff R rR;
         tp' \<leftarrow> rule_removal_tt I (get_redtriple redp) Rr tp 
           <+? (\<lambda>s. i +@+ '': error when applying the rule removal technique on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+
             '' to remove the rules '' +#+ shows_nl +@+ shows_rules Rr +@+ shows_nl +@+ s);
         check_trs_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below the rule removal technique'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_trs_proof a i I J tp (DPTrans P prf) = debug (i []) ''DP trans'' (do {
     dpp \<leftarrow> dependency_pairs_tt I J tp Sharp P
       <+? (\<lambda>s. i +@+ '': error when switching from the TRS'' +#+ shows_nl +@+ shows_tp I tp
         +@+ shows_nl +@+ ''to the initial DP problem with pairs '' +#+ shows_nl +@+ shows_rules P +@+ shows_nl +@+ s);
     check_dp_proof a (i +@+ shows_string ''.1'') J dpp prf
       <+? (\<lambda>s. i +@+ '': error below switch to dependency pairs'' +#+ shows_nl +@+ indent s)
   })" 
|  "check_trs_proof a i I J tp (StringReversal prf) = debug (i []) ''String Reversal'' (do {
         tp' \<leftarrow> string_reversal_tt I tp 
           <+? (\<lambda>s. i +@+ '': error when applying string reversal on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s);
         check_trs_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below the string reversal technique'' +#+ shows_nl
             +@+ indent s)
   })" 
|  "check_trs_proof a i I J tp (Semlab sli lQ lAll prf) = debug (i []) ''Semlab'' (do {
         tp' \<leftarrow> semlab_fin_tt I sli lQ lAll tp 
           <+? (\<lambda>s. i +@+ '': error when applying semantic labelling on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s);
         check_trs_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below the sem.lab technique'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_trs_proof a i I J tp (Bounds info) = debug (i []) ''Bounds'' (do {
     bounds_tt I info tp 
       <+? (\<lambda>s. i +@+ '': error when applying bounds on the termination problem '' +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s)})" 
| "check_trs_proof a i I J tp (Uncurry u_info rR prf) = debug (i []) ''Uncurry'' (do {
     tp' \<leftarrow> uncurry_tt I u_info rR tp
       <+? (\<lambda>s. i +@+ '': error when applying currying transformation'' +#+ shows_nl +@+ s);
     check_trs_proof a (i +@+ shows_string ''.1'') I J tp' prf
       <+? (\<lambda>s. i +@+ '': error below the uncurrying technique'' +#+ shows_nl +@+ indent s)
   })"
| "check_trs_proof a i I J tp (Fcc fcs R' prf) = debug (i []) ''Fcc'' (do {
    tp' \<leftarrow> fcc_tt I fcs R' tp
      <+? (\<lambda>s. i +@+ '': error when applying flat context closure'' +#+ shows_nl +@+ s);
    check_trs_proof a (i +@+ shows_string ''.1'') I J tp' prf
      <+? (\<lambda>s. i +@+ '': error below flat context closure'' +#+ shows_nl +@+ indent s)
  })"
| "check_trs_proof a i I J tp (Split Rrem prf1 prf2) = debug (i []) ''Split'' (do {
    let (tp1,tp2) = split_tt I tp Rrem;
    check_trs_proof a (i +@+ shows_string ''.1'') I J tp1 prf1
      <+? (\<lambda>s. i +@+ '': error below the splitting'' +#+ shows_nl
        +@+ indent s);
    check_trs_proof a (i +@+ shows_string ''.2'') I J tp2 prf2
      <+? (\<lambda>s. i +@+ '': error below the splitting'' +#+ shows_nl
        +@+ indent s)
  })"
|  "check_trs_proof a i I J tp (SwitchInnermost joins prf) = debug (i []) ''Switch Innermost'' (do {
         tp' \<leftarrow> switch_innermost_tt I joins tp 
           <+? (\<lambda>s. i +@+ '': error when switching to innermost on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s);
         check_trs_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below the switch to innermost'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_trs_proof a i I J tp (AssumeSN tp') =
    debug (i []) ''Termination Assumption'' (
      if a
        then check_tp_subsumes I tp' tp
          <+? (\<lambda>s. i +@+ '': error in termination assumption'' +#+ shows_nl +@+ s +@+ shows_nl)
        else error (i +@+ '': the proof contains a local termination assumption'' +#+ shows_nl)
    )"



fun assms :: "bool \<Rightarrow> ('f, 'l, 'v) trs_proof \<Rightarrow> ('f, 'l, 'v) prob list" where
  "assms a (DPTrans _ p) = map Inr (dp_assms a p)"
| "assms a (RuleRemoval _ _ p) = assms a p"
| "assms a (StringReversal p) = assms a p"
| "assms a (Bounds _) = []"
| "assms a (Uncurry _ _ p) = assms a p"
| "assms a (Semlab _ _ _ p) = assms a p"
| "assms a RisEmpty = []"
| "assms a (Fcc _ _ p) = assms a p"
| "assms a (Split _ p q) = assms a p @ assms a q"
| "assms a (SwitchInnermost _ p) = assms a p"
| "assms a (AssumeSN tp) = (if a then [Inl tp] else [])"

lemma assms_False[simp]: "assms False prf = []"
  by (induct False "prf" rule: assms.induct) simp_all

fun holds :: "('f, 'l, 'v) prob \<Rightarrow> bool" where
  "holds (Inl p) = SN_qrelL p"
| "holds (Inr p) = finite_dppL p"

lemma check_trs_proof_with_assms_sound:
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and fin: "\<forall>p\<in>set (assms a prf). holds p"
  and ok: "isOK (check_trs_proof a i I J tp prf)"
  shows "SN_qrel (tp_ops.qreltrs I tp)"
proof -
  interpret tp_spec I by fact
  from ok fin show ?thesis
  proof (induct "prf" arbitrary: tp i)
    case (RisEmpty i)
    thus ?case by (auto simp: SN_qrel_def rel_SN_def rel_step.simps)
  next
    case (RuleRemoval rp rR prof tp i)
    then obtain tp' i' Rr where tt: "rule_removal_tt I (get_redtriple rp) Rr tp = return tp'"
    and rec: "isOK (check_trs_proof a i' I J tp' prof)"
      by (auto simp: Let_def)
    from RuleRemoval have fin: "\<forall>p\<in>set (assms a prof). holds p" by simp
    show ?case
      by (rule sound_tt_impl[OF rule_removal_tt[OF get_redtriple],
        OF tt RuleRemoval(1)[OF rec fin]])
  next
    case (DPTrans P prof tp i)
    then obtain dpp i' shp where tt: "dependency_pairs_tt I J tp shp P = return dpp" 
      and rec: "isOK (check_dp_proof a i' J dpp prof)" by auto
    from DPTrans have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
    show ?case
      by (rule dependency_pairs_tt[OF I J tt
        check_dp_proof_with_dp_assms[OF J fin rec]])
  next
    case (StringReversal prof tp i)
    then obtain tp' i' where tt: "string_reversal_tt I tp = return tp'"
      and rec: "isOK (check_trs_proof a i' I J tp' prof)"
      by (auto simp: Let_def)
    from StringReversal have fin: "\<forall>p\<in>set (assms a prof). holds p" by simp
    show ?case
      by (rule sound_tt_impl[OF string_reversal_tt, OF tt StringReversal(1)[OF rec fin]])
  next
    case (Semlab sli lQ lAll prof tp i)
    then obtain tp' i' where tt: "semlab_fin_tt I sli lQ lAll tp = return tp'"
    and rec: "isOK(check_trs_proof a i' I J tp' prof)"
      by (auto simp: Let_def)
    from Semlab have fin: "\<forall>p\<in>set (assms a prof). holds p" by simp
    show ?case
      by (rule sound_tt_impl[OF semlab_fin_tt[OF I], OF tt Semlab(1)[OF rec fin]])
  next
    case (Bounds info)
    hence tt: "isOK (bounds_tt I info tp)"
      by (auto simp: Let_def)
    show ?case
      by (rule bounds_tt[OF I tt])
  next
    case (Uncurry mode info prof tp i)
    then obtain tp' i' where tt: "uncurry_tt I mode info tp = return tp'"
      and rec: "isOK (check_trs_proof a i' I J tp' prof)"
      by (auto simp: Let_def)
    from Uncurry have fin: "\<forall>p\<in>set (assms a prof). holds p" by simp
    show ?case
      by (rule sound_tt_impl[OF uncurry_tt, OF _ tt Uncurry(1)[OF rec fin]], unfold_locales)
  next
    case (Fcc fcs R' prof tp i)
    then obtain tp' i' where tt: "fcc_tt I fcs R' tp = return tp'"
      and rec: "isOK (check_trs_proof a i' I J tp' prof)" by auto
    from Fcc have fin: "\<forall>p\<in>set (assms a prof). holds p" by simp
    from Fcc(1)[OF rec fin] have "SN_qrel (tp_ops.qreltrs I tp')" .
    from sound_tt_impl[OF fcc_tt, OF tt this] show ?case .
  next
    case (Split R' prof1 prof2 tp i)
    let ?i = "\<lambda>s. i +@+ shows_string s"
    let ?i1 = "?i ''.1''"
    let ?i2 = "?i ''.2''"
    obtain tp1 tp2 where s: "split_tt I tp R' = (tp1,tp2)" by force
    from Split(3) s have ok1: "isOK (check_trs_proof a ?i1 I J tp1 prof1)" 
      and ok2: "isOK (check_trs_proof a ?i2 I J tp2 prof2)" by auto
    from Split have fin1: "\<forall>p\<in>set (assms a prof1). holds p"
      and fin2: "\<forall>p\<in>set (assms a prof2). holds p" by simp_all
    show ?case
      by (rule split_tt[OF s Split(1)[OF ok1 fin1] Split(2)[OF ok2 fin2]])
  next
    case (SwitchInnermost joins prof tp i)
    then obtain tp' i' where tt: "switch_innermost_tt I joins tp = return tp'"
      and rec: "isOK (check_trs_proof a i' I J tp' prof)"
      by (auto simp: Let_def)
    from SwitchInnermost have fin: "\<forall>p\<in>set (assms a prof). holds p" by simp
    show ?case
      by (rule sound_tt_impl[OF switch_innermost_tt[OF I], OF infinite_lab tt SwitchInnermost(1)[OF rec fin]]) 
  next
    case (AssumeSN tp' tp i)
    have "a = True \<or> a = False" by simp
    thus ?case
    proof
      assume False: "a = False" thus ?thesis using AssumeSN(1) by simp
    next
      assume True: "a = True"
      show ?thesis
      proof (cases tp')
        case fields
        show ?thesis using AssumeSN[unfolded True, simplified] and check_tp_subsumes_sound
          and `tp_spec I`
          by best
      qed
    qed
  qed
qed

lemma check_trs_proof_sound:
  assumes "tp_spec I" and "dpp_spec J"
    and "isOK (check_trs_proof False i I J tp prf)"
  shows "SN_qrel (tp_ops.qreltrs I tp)"
  using check_trs_proof_with_assms_sound[OF assms(1,2) _ assms(3)] by simp

end