(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Generic_Reduction_Pair_Processor_Impl
imports
  Generic_Usable_Rules_Impl
  Generic_Reduction_Pair_Processor
begin

definition
  generic_ur_af_redtriple_proc ::
    "('dpp, 'f, string) dpp_ops \<Rightarrow> ('f::{show,linorder}, string) redtriple \<Rightarrow> 
     ('f,string)rules option \<Rightarrow> ('f,string)rules \<Rightarrow> 'dpp proc"
where
  "generic_ur_af_redtriple_proc I rp U_opt Premove dpp = check_return (do {
     redtriple.valid rp;
     let (ps, pns) = dpp_ops.split_pairs I dpp Premove;
     let P = dpp_ops.pairs I dpp;
     U \<leftarrow> smart_usable_rules_checker_impl I dpp (redtriple.af rp) U_opt P;
     check_allm (redtriple.ns rp) U
       <+? (\<lambda>s. ''problem when orienting (usable) rules'' +#+ shows_nl +@+ s);
     check_allm (redtriple.nst rp) pns
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s);
     check_allm (redtriple.s rp) ps
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
   } <+? (\<lambda>s. ''could not apply the generic reduction pair processor with the following'' +#+ shows_nl +@+
     (redtriple.desc rp) +@+ shows_nl +@+ s))
   (dpp_spec.delete_pairs I dpp Premove)"

lemma generic_ur_af_redtriple_proc:
  assumes I: "dpp_spec I"
  and grp: "generic_redtriple_impl grp"
  shows "dpp_spec.sound_proc_impl I (generic_ur_af_redtriple_proc I (grp rp) U_opt ps)"
proof -
  interpret dpp_spec I by fact
  interpret generic_redtriple_impl grp by fact
  show ?thesis
  proof
    fix d d'
    assume fin: "finite_dpp (dpp d')"
      and ok: "generic_ur_af_redtriple_proc I (grp rp) U_opt ps d = return d'"
    let ?rp = "grp rp"
    let ?S = "set ps"
    let ?P = "set (P d)"
    let ?Pw = "set (Pw d)"
    let ?Pb = "set (pairs d)"
    let ?PP = "?P \<union> ?Pw"
    let ?Q = "set (Q d)"
    let ?R = "set (R d)"
    let ?Rw = "set (Rw d)"
    let ?Rb = "set (rules d)"
    let ?RR = "set (R d @ Rw d)"
    have RR_conv: "?RR = ?Rb" unfolding dpp_spec_sound by auto
    have PP_conv: "?PP = ?Pb" unfolding dpp_spec_sound by auto
    show "finite_dpp (dpp d)"
    proof (rule finite_under_var_cond_impl)
      assume var: "\<And>lr. lr \<in> ?Rb \<Longrightarrow> is_Fun (fst lr)"
      show ?thesis
      proof (cases "split_pairs d ps")
        case (Pair Ps Pr)    
        from split_pairs_sound[OF this] have Ps: "set Ps = (?P \<union> ?Pw) \<inter> ?S" 
          and Pr: "set Pr = (?P \<union> ?Pw) - ?S" by (auto simp: dpp_spec_sound)
        let ?pi = "redtriple.af ?rp"
        let ?urcheck = "smart_usable_rules_checker_impl I d ?pi U_opt (pairs d)" 
        note ok = ok[unfolded generic_ur_af_redtriple_proc_def Let_def Pair, simplified]
        from ok have valid: "isOK(redtriple.valid ?rp)" 
          and compat: "isOK(?urcheck)" by auto 
        from compat obtain U where urc: "?urcheck = return U" by (cases ?urcheck, auto) 
        note ok = ok[unfolded urc]
        from ok have
              NS: "isOK (check_allm (redtriple.ns ?rp) U)"
          and NST: "isOK (check_allm (redtriple.nst ?rp) Pr)"
          and S: "isOK(check_allm (redtriple.s ?rp) Ps)" 
          and d': "d' = delete_pairs d ps" by auto
        from generate_redtriple[OF valid S NS NST, unfolded Ps] 
        obtain S NS NST where "ce_af_redtriple_order S NS NST (redtriple.af ?rp)" and S: "(?P \<union> ?Pw) \<inter> ?S \<subseteq> S" and NS: "?P \<union> ?Pw - ?S \<subseteq> NST" and R: "set U \<subseteq> NS" by  (auto simp: dpp_spec_sound Pr)
        hence NS: "?P \<union> ?Pw \<subseteq> NST \<union> S" by auto
        from smart_usable_rules_checker_impl[OF I urc var] have ur: "smart_usable_rules_checker ?pi ?Q (rules d) U_opt ?PP = Some U" by auto
        interpret ce_af_redtriple_order S NS NST ?pi by fact
        have ce: "ce_af_redtriple S NS NST ?pi" ..
        note proc = generic_redtriple_proc[OF smart_usable_rules_checker R NS ce, OF ur]
        show ?thesis
          unfolding finite_dpp_def
        proof(rule, elim exE) 
          fix s t \<sigma>
          assume min: "min_ichain (dpp d) s t \<sigma>"
          have "\<exists> i. min_ichain (?P - ?S,?Pw - ?S,?Q,?R,?Rw) (shift s i) (shift t i) (shift \<sigma> i)" 
            by (rule min_ichain_split[OF min[unfolded dpp_sound], of ?S "{}", unfolded Int_empty_left Diff_empty],
              rule rev_contrapos[OF _ proc[of s t \<sigma>, unfolded dpp_spec_sound]], rule min_ichain_mono[of "?S \<inter> ?PP" "?PP - ?S" ?Q "{}" "?R \<union> ?Rw"], insert S, auto 
              )
          with fin[unfolded d' delete_P_Pw_sound finite_dpp_def] show False by blast
        qed
      qed
    qed
  qed
qed

definition root_aft_to_entry where
  "root_aft_to_entry s t \<pi> \<equiv> let rt = the (root t); \<pi>t = \<pi> rt; ts = args t
        in map (\<lambda> i. (s, ts ! i)) (filter (\<lambda> i. \<pi>t i) [0 ..< snd rt])"

definition
  generic_ur_af_root_redtriple_proc ::
    "('dpp, 'f, string) dpp_ops \<Rightarrow> ('f::{show,linorder}, string) root_redtriple \<Rightarrow> 
     ('f,string)rules option \<Rightarrow> ('f,string)rules \<Rightarrow> 'dpp proc"
where
  "generic_ur_af_root_redtriple_proc I rp U_opt Premove dpp = check_return (do {
     root_redtriple.valid rp;
     let (ps, pns) = dpp_ops.split_pairs I dpp Premove;
     let P = dpp_ops.pairs I dpp;
     let pi = root_redtriple.af rp;
     let pi' = root_redtriple.aft rp;
     let is_def = dpp_spec.is_defined I dpp;
     check_allm (\<lambda>(l, r). do {
        check_no_var l;
        check_no_var r;
        check_no_defined_root is_def r
      }) P;
     U \<leftarrow> smart_usable_rules_checker_impl I dpp pi U_opt (concat (map (\<lambda> (s,t). root_aft_to_entry s t pi') P));
     check_allm (root_redtriple.ns rp) U
       <+? (\<lambda>s. ''problem when orienting (usable) rules'' +#+ shows_nl +@+ s);
     check_allm (root_redtriple.nst rp) pns
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s);
     check_allm (root_redtriple.s rp) ps
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
   } <+? (\<lambda>s. ''could not apply the generic root reduction pair processor with the following'' +#+ shows_nl +@+
     (root_redtriple.desc rp) +@+ shows_nl +@+ s))
   (dpp_spec.delete_pairs I dpp Premove)"

lemma generic_ur_af_root_redtriple_proc:
  assumes I: "dpp_spec I"
  and grp: "generic_root_redtriple_impl grp"
  shows "dpp_spec.sound_proc_impl I (generic_ur_af_root_redtriple_proc I (grp rp) U_opt ps)"
proof -
  interpret dpp_spec I by fact
  interpret generic_root_redtriple_impl grp by fact
  show ?thesis
  proof
    fix d d'
    assume fin: "finite_dpp (dpp d')"
      and ok: "generic_ur_af_root_redtriple_proc I (grp rp) U_opt ps d = return d'"
    let ?rp = "grp rp"
    let ?S = "set ps"
    let ?P = "set (P d)"
    let ?Pw = "set (Pw d)"
    let ?Pb = "set (pairs d)"
    let ?PP = "?P \<union> ?Pw"
    let ?Q = "set (Q d)"
    let ?R = "set (R d)"
    let ?Rw = "set (Rw d)"
    let ?Rb = "set (rules d)"
    let ?RR = "set (R d @ Rw d)"
    have RR_conv: "?RR = ?Rb" unfolding dpp_spec_sound by auto
    have PP_conv: "?PP = ?Pb" unfolding dpp_spec_sound by auto
    show "finite_dpp (dpp d)"
    proof (rule finite_under_var_cond_impl)
      assume var: "\<And>lr. lr \<in> ?Rb \<Longrightarrow> is_Fun (fst lr)"
      show ?thesis
      proof (cases "split_pairs d ps")
        case (Pair Ps Pr)    
        from split_pairs_sound[OF this] have Ps: "set Ps = (?P \<union> ?Pw) \<inter> ?S" 
          and Pr: "set Pr = (?P \<union> ?Pw) - ?S" by (auto simp: dpp_spec_sound)
        let ?pi = "root_redtriple.af ?rp"
        let ?pi' = "root_redtriple.aft ?rp"
        let ?P' = "concat (map (\<lambda> (s,t). root_aft_to_entry s t ?pi') (pairs d))"
        let ?R = "rules d"
        let ?urcheck = "smart_usable_rules_checker_impl I d ?pi U_opt ?P'" 
        note ok = ok[unfolded generic_ur_af_root_redtriple_proc_def Let_def Pair, simplified]
        from ok have valid: "isOK(root_redtriple.valid ?rp)" 
          and compat: "isOK(?urcheck)" by auto 
        from compat obtain U where urc: "?urcheck = return U" by (cases ?urcheck, auto) 
        note ok = ok[unfolded urc]
        from ok have
              NS: "isOK (check_allm (root_redtriple.ns ?rp) U)"
          and NST: "isOK (check_allm (root_redtriple.nst ?rp) Pr)"
          and S: "isOK(check_allm (root_redtriple.s ?rp) Ps)" 
          and d': "d' = delete_pairs d ps" 
          and Pcond: "\<And> s t. (s,t) \<in> ?P \<union> ?Pw \<Longrightarrow> is_Fun s \<and> is_Fun t" 
          and ndef: "\<forall> (s,t) \<in> ?P \<union> ?Pw. \<not> defined (set ?R) (the (root t))" unfolding check_no_defined_root_sound[OF refl, of ?R, simplified] by auto
        from generate_root_redtriple[OF valid S NS NST, unfolded Ps] 
        obtain S NS NST where "ce_af_root_redtriple_order S NS NST ?pi ?pi'" and S: "(?P \<union> ?Pw) \<inter> ?S \<subseteq> S" and NS: "?P \<union> ?Pw - ?S \<subseteq> NST" and R: "set U \<subseteq> NS" by  (auto simp: dpp_spec_sound Pr)
        hence NS: "?P \<union> ?Pw \<subseteq> NST \<union> S" by auto
        from smart_usable_rules_checker_impl[OF I urc var] have ur: "smart_usable_rules_checker ?pi ?Q (rules d) U_opt (set ?P') = Some U" by auto
        interpret ce_af_root_redtriple_order S NS NST ?pi ?pi' by fact
        have ce: "ce_af_root_redtriple S NS NST ?pi ?pi'" ..
        note proc = root_redtriple_sound[OF smart_usable_rules_checker _ R NS Pcond ndef ce, of ?Q U_opt] 
        let ?PP = "?P \<union> ?Pw"
        have no_chain: "\<And> s t \<sigma>. \<not> min_ichain (S \<inter> ?PP, ?PP - S, ?Q, {}, set ?R) s t \<sigma>"
        proof (rule proc, unfold ur[symmetric], rule arg_cong[where f = "smart_usable_rules_checker ?pi ?Q (rules d) U_opt"])
          show "{u. \<exists> s f ts i. u = (s, ts ! i) \<and> (s,Fun f ts) \<in> ?P \<union> ?Pw \<and> i < length ts \<and> ?pi' (f,length ts) i} = set ?P'"
            (is "?one = ?two")
          proof -
            {
              fix s f ts i
              assume "(s,Fun f ts) \<in> ?P \<union> ?Pw \<and> i < length ts \<and> ?pi' (f,length ts) i"
              hence "(s, ts ! i) \<in> ?two" unfolding root_aft_to_entry_def Let_def unfolding pairs_sound[symmetric] by force
            }
            hence one: "?one \<subseteq> ?two" by blast
            {
              fix s ti
              assume "(s,ti) \<in> ?two"
              then obtain s' t' where st': "(s',t') \<in> set (pairs d)" 
                and mem: "(s,ti) \<in> set (root_aft_to_entry s' t' ?pi')" by auto
              from Pcond[of s' t'] st' obtain f ts where t': "t' = Fun f ts" by (cases t', auto)
              from mem[unfolded t' root_aft_to_entry_def Let_def, simplified]
              obtain i where ti: "ti = ts ! i" and i: "i < length ts" and pi: "?pi' (f,length ts) i" and s: "s = s'"
                by auto
              have "(s,ti) \<in> ?one" unfolding ti s
                by (rule, intro exI conjI, rule refl, insert st' pi i t', auto)
            }
            hence two: "?two \<subseteq> ?one" ..
            from one two show ?thesis by blast
          qed
        qed          
        show ?thesis
          unfolding finite_dpp_def
        proof(rule, elim exE) 
          fix s t \<sigma>
          assume min: "min_ichain (dpp d) s t \<sigma>"
          have "\<exists> i. min_ichain (?P - ?S,?Pw - ?S,?Q,set (R d),set (Rw d)) (shift s i) (shift t i) (shift \<sigma> i)" 
            by (rule min_ichain_split[OF min[unfolded dpp_sound], of ?S "{}", unfolded Int_empty_left Diff_empty],
            rule rev_contrapos[OF _ no_chain[of s t \<sigma>]], rule min_ichain_mono[of "?S \<inter> ?PP" "?PP - ?S" ?Q "{}" "set ?R"], 
              insert S, auto)              
          with fin[unfolded d' delete_P_Pw_sound finite_dpp_def] show False by blast
        qed
      qed
    qed
  qed
qed


end
