(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory "Q-Restricted_Rewriting_Impl"
imports
  "Q-Restricted_Rewriting"
  Trs_Impl
  Innermost_Loops_Impl
begin


definition prop_rstep_r_p_s where "prop_rstep_r_p_s P R r p \<sigma> \<equiv> {(s,t). (\<forall> u \<in> set (args (s |_ p)). P u) \<and>
  p \<in> poss s \<and> r \<in> R \<and> s |_ p = fst r \<cdot> \<sigma> \<and> t = replace_at s p (snd r \<cdot> \<sigma>)}"

definition
  check_prop_rstep_rule ::
    "(('f::show, 'v::show) term \<Rightarrow> shows check) \<Rightarrow>
     pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow>
     shows check"
where
  "check_prop_rstep_rule P p rule s t \<equiv> do {
     check (in_poss p s) (shows p +@+ '' is not a position of '' +#+ shows s +@+ shows_nl);
     check (in_poss p t) (shows p +@+ '' is not a position of '' +#+ shows t +@+ shows_nl);
     let C = ctxt_of_pos_term p s;
     let D = ctxt_of_pos_term p t;
     let u = subt_at s p;
     let v = subt_at t p;
     (case match_list u [(u, fst rule), (v, snd rule)] of
       Some \<sigma> \<Rightarrow>
         check_allm P (args u)
       \<guillemotright> check (C = D)
           (''the term '' +#+ shows t
             +@+ '' does not result from a proper application of rule ''
             +#+ shows_nl +@+ shows_rule rule +@+ '' at position '' +#+ shows p +@+ shows_nl)
     | None \<Rightarrow>
       error (''the term '' +#+ shows t
             +@+ '' does not result from a proper application of rule ''
             +#+ shows_nl +@+ shows_rule rule +@+ '' at position '' +#+ shows p +@+ shows_nl))
  }"

lemma check_prop_rstep_rule[simp]:
  "isOK (check_prop_rstep_rule P p r s t) = (\<exists>\<sigma>. (s, t) \<in> prop_rstep_r_p_s (\<lambda> t. isOK(P t)) {r} r p \<sigma>)" (is "?l = ?r")
proof 
  assume ok: ?l
  let ?C = "ctxt_of_pos_term p s"
  let ?D = "ctxt_of_pos_term p t"
  let ?mlist = "[(subt_at s p, fst r), (subt_at t p, snd r)]"
  note ok = ok[unfolded check_prop_rstep_rule_def Let_def, simplified]
  from ok have "p \<in> poss s" and "p \<in> poss t" by auto
  from ok obtain \<sigma> where some: "match_list (subt_at s p) [(subt_at s p, fst r), (subt_at t p, snd r)] = Some \<sigma>"
    by force
  from match_list_sound[OF this] have "fst r \<cdot> \<sigma> = s |_ p" and "snd r \<cdot> \<sigma> = t |_ p" by auto
  hence "s |_ p = fst r \<cdot> \<sigma>" and "t = ?D\<langle>snd r \<cdot> \<sigma>\<rangle>"
    using ctxt_supt_id[OF `p \<in> poss t`]  by simp+
  moreover from ok and some
    have "?D = ?C" by simp
  moreover from ok and some
  have "\<And> t. t \<in> set (args (s |_ p)) \<Longrightarrow> isOK( P t )" by simp
  ultimately have "(s, t) \<in> prop_rstep_r_p_s (\<lambda> t. isOK (P t)) {r} r p \<sigma>"
    using `p \<in> poss s` unfolding prop_rstep_r_p_s_def some by force
  thus ?r by blast
next
  assume r: ?r  
  let ?P = "\<lambda> t. isOK (P t)"
  from r obtain \<sigma> where step: "(s, t) \<in> prop_rstep_r_p_s ?P {r} r p \<sigma>" by auto
  let ?C = "ctxt_of_pos_term p s"
  let ?D = "ctxt_of_pos_term p t"
  from step have "p \<in> poss s" 
    and P: "\<forall>u \<in> set (args (s |_ p)). ?P u"
    and s: "s |_ p = fst r \<cdot> \<sigma>" and t: "t = ?C\<langle>snd r \<cdot> \<sigma>\<rangle>"
    unfolding prop_rstep_r_p_s_def by auto
  from hole_pos_ctxt_of_pos_term[OF `p \<in> poss s`]
  have p: "p = hole_pos ?C" by simp
  have "p \<in> set (poss_list s)" by (simp add: `p \<in> poss s`)
  moreover have "p \<in> set (poss_list t)"
    using p and hole_pos_poss[of "?C" "snd r \<cdot> \<sigma>"] by (simp add: t)
  moreover obtain \<tau> where "match_list (s |_ p) [(s |_ p, fst r), (t |_ p, snd r)] = Some \<tau>"
    and tau: "\<forall>x\<in>rvars [(s |_ p, fst r), (t |_ p, snd r)]. \<tau> x = \<sigma> x"
  proof -
    from subt_at_hole_pos[of "?C" "snd r \<cdot> \<sigma>", unfolded t[symmetric]]
    have 2: "t |_ p = snd r \<cdot> \<sigma>" unfolding p[symmetric] .
    assume "\<And>\<tau>. \<lbrakk>match_list (s |_ p) [(s |_ p, fst r), (t |_ p, snd r)] = Some \<tau>;
      \<forall>x\<in>rvars [(s |_ p, fst r), (t |_ p, snd r)]. \<tau> x = \<sigma> x\<rbrakk> \<Longrightarrow> thesis"
    thus ?thesis using match_list_complete unfolding s 2 by force
  qed
  moreover have "isOK(check_allm P (args (s |_ p)))" using P by auto
  moreover have "?C = ?D"
  proof -
    from hole_pos_id_ctxt[OF t[symmetric]]
    show ?thesis unfolding p[symmetric] ..
  qed
  ultimately show ?l unfolding check_prop_rstep_rule_def Let_def by auto
qed


definition
  check_prop_rstep ::
    "(('f::show, 'v::show) term \<Rightarrow> shows check) \<Rightarrow>
     ('f, 'v) rules \<Rightarrow>
     pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow>
     shows check"
where
  "check_prop_rstep P R p rule s t \<equiv> do {
     check (\<exists> r \<in> set R. rule =v r)
       (shows_rule rule +@+ '' is not a rule of'' +#+ shows_nl +@+ shows_trs R +@+ shows_nl);
     check_prop_rstep_rule P p rule s t
   }"

lemma check_prop_rstep_sound:
  assumes ok: "isOK (check_prop_rstep P R p r s t)"
  shows "\<exists>\<sigma> r'. (s, t) \<in> prop_rstep_r_p_s (\<lambda> t. isOK(P t)) (set R) r' p \<sigma> \<and> r =v r'"
proof -
  note ok = ok[unfolded check_prop_rstep_def, simplified]
  from ok obtain rule where rule: "rule \<in> set R" and inst: "r =v rule" by auto
  from inst obtain \<delta> where fst: "fst r = fst rule \<cdot> \<delta>" and snd: "snd r = snd rule \<cdot> \<delta>"
    unfolding eq_rule_mod_vars_def instance_rule_def by auto
  from ok obtain \<sigma> where step: "(s,t) \<in> prop_rstep_r_p_s (\<lambda> t. isOK(P t)) {r} r p \<sigma>" by auto
  show ?thesis
  proof (intro exI conjI)
    show "(s,t) \<in> prop_rstep_r_p_s (\<lambda> t. isOK(P t)) (set R) rule p (\<delta> \<circ>s \<sigma>)" using
      step unfolding prop_rstep_r_p_s_def using rule fst snd by simp
  qed (rule inst)
qed


lemma check_prop_rstep_complete:
  assumes "\<exists>\<sigma>. (s, t) \<in> prop_rstep_r_p_s (\<lambda> t. isOK (P t)) (set R) r p \<sigma>"
  shows "isOK (check_prop_rstep P R p r s t)"
proof -
  let ?P = "\<lambda> t. isOK (P t)"
  from assms obtain \<sigma> where step: "(s, t) \<in> prop_rstep_r_p_s ?P (set R) r p \<sigma>" by auto
  from step have step: "(s, t) \<in> prop_rstep_r_p_s ?P {r} r p \<sigma>" and rule: "r \<in> set R" 
    unfolding prop_rstep_r_p_s_def by auto
  have "r =v r" by simp
  with rule have "\<exists> rule \<in> set R. r =v rule" by blast
  with step show ?thesis unfolding check_prop_rstep_def by auto
qed

definition
  check_prop_rstep' ::
    "(('f::show, 'v::show) term \<Rightarrow> shows check) \<Rightarrow>
     ('f, 'v) rules \<Rightarrow>
     pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow>
     shows check"
where
  "check_prop_rstep' P R p rule s t \<equiv> do {
     check (rule \<in> set R)
       (shows_rule rule +@+ '' is not a rule of'' +#+ shows_nl +@+ shows_trs R +@+ shows_nl);
     check_prop_rstep_rule P p rule s t
   }"

lemma check_prop_rstep'[simp]:
  "isOK (check_prop_rstep' P R p r s t) = 
  (\<exists>\<sigma>. (s, t) \<in> prop_rstep_r_p_s (\<lambda> t. isOK(P t)) (set R) r p \<sigma>)"
  unfolding check_prop_rstep'_def
  by (auto simp: prop_rstep_r_p_s_def)

lemma prop_rstep_qrstep: "prop_rstep_r_p_s (\<lambda> t. t \<in> NF_terms Q) = qrstep_r_p_s Q" (is "?l = ?r")
proof (rule ext, rule ext, rule ext, rule ext)
  fix R r p \<sigma>
  show "?l R r p \<sigma> = ?r R r p \<sigma>"
  unfolding prop_rstep_r_p_s_def qrstep_r_p_s_def NF_terms_args_conv[symmetric]
  using subt_at_ctxt_of_pos_term[of _ p "fst r \<cdot> \<sigma>"]
  by auto
qed

lemma prop_rstep_rstep: "prop_rstep_r_p_s (\<lambda> t. True) = rstep_r_p_s" (is "?l = ?r")
  by (intro ext, unfold prop_rstep_r_p_s_def rstep_r_p_s_def', simp)

lemma prop_rstep_qrstep_subst: fixes \<mu> :: "('f,'v)subst"
  assumes nvar: "is_Fun (fst r)"
  shows "(s,t) \<in> prop_rstep_r_p_s (\<lambda> t. (\<forall> i. t \<cdot> \<mu> ^^ i \<in> NF_terms Q)) R r p \<sigma> = (\<forall> i. (s \<cdot> \<mu> ^^ i, t \<cdot> \<mu> ^^ i) \<in> qrstep_r_p_s Q R r p (\<sigma> \<circ>s \<mu> ^^ i))" (is "?l = ?r")
proof -
  let ?NF = "\<lambda> t i. (\<forall> i. t \<cdot> \<mu> ^^ i \<in> NF_terms Q)"
  let ?c = "ctxt_of_pos_term p s"
  let ?QR = "qrstep_r_p_s Q R r p"
  from nvar obtain f ls where l: "fst r = Fun f ls" by force
  show ?thesis
  proof
    assume ?l
    from this[unfolded prop_rstep_r_p_s_def Let_def]
    have p: "p \<in> poss s" and r: "r \<in> R" and s: "s |_ p = fst r \<cdot> \<sigma>" and t: "t = ?c\<langle>snd r \<cdot> \<sigma>\<rangle>" 
      and NF: "\<And> u i. u \<in> set (args s |_ p) \<Longrightarrow> u \<cdot> \<mu> ^^ i \<in> NF_terms Q" by auto
    then obtain us where sp: "s |_ p = Fun f us" unfolding l by (cases "s |_ p", auto)
    from p have p': "\<And> i. p \<in> poss (s \<cdot> \<mu> ^^ i)" by (rule poss_imp_subst_poss) 
    show ?r 
    proof
      fix i
      let ?C = "ctxt_of_pos_term p (s \<cdot> \<mu> ^^ i)"
      show "(s \<cdot> \<mu> ^^ i, t \<cdot> \<mu> ^^ i) \<in> ?QR (\<sigma> \<circ>s \<mu> ^^ i)"
        unfolding prop_rstep_qrstep[symmetric] prop_rstep_r_p_s_def Let_def
      proof (rule, rule, intro conjI ballI)
        fix u
        assume "u \<in> set (args s \<cdot> \<mu> ^^ i |_ p)"
        then obtain v where v: "v \<in> set (args s |_ p)" and u: "u = v \<cdot> \<mu> ^^ i" unfolding subt_at_subst[OF p] sp
          by auto        
        show "u \<in> NF_terms Q" unfolding u using NF[OF v] .
      next
        show "p \<in> poss (s \<cdot> \<mu> ^^ i)" by (rule p')
      next
        show "r \<in> R" by (rule r)
      next
        have "fst r \<cdot> \<sigma> \<circ>s \<mu> ^^ i = s |_ p \<cdot> \<mu> ^^ i" using s[symmetric] by simp
        also have "... = s \<cdot> \<mu> ^^ i |_ p" unfolding subt_at_subst[OF p] ..
        finally show "s \<cdot> \<mu> ^^ i |_ p = fst r \<cdot> \<sigma> \<circ>s \<mu> ^^ i" by simp
      next          
        show "t \<cdot> \<mu> ^^ i = ?C\<langle>snd r \<cdot> \<sigma> \<circ>s \<mu> ^^ i\<rangle>"
          unfolding t subst_apply_term_ctxt_apply_distrib ctxt_of_pos_term_subst[OF p] ctxt_eq by simp
      qed
    qed
  next
    assume ?r
    hence step: "\<And> i. (s \<cdot> \<mu> ^^ i, t \<cdot> \<mu> ^^ i) \<in> prop_rstep_r_p_s (\<lambda> t. t \<in> NF_terms Q) R r p (\<sigma> \<circ>s \<mu> ^^ i)"
      unfolding prop_rstep_qrstep ..
    note step = step[unfolded prop_rstep_r_p_s_def Let_def]
    from step[of 0]
    have p: "p \<in> poss s" and r: "r \<in> R" and s: "s |_ p = fst r \<cdot> \<sigma>" and t: "t = ?c\<langle>snd r \<cdot> \<sigma>\<rangle>" by auto
    hence rs: "fst r \<cdot> \<sigma> = s |_ p" by simp
    then obtain us where sp: "s |_ p = Fun f us" unfolding l by (cases "s |_ p", auto)
    show ?l
      unfolding prop_rstep_r_p_s_def Let_def
    proof (rule, rule, intro conjI ballI)
      fix u
      assume u: "u \<in> set (args s |_ p)"
      show "\<forall> i. u \<cdot> \<mu> ^^ i \<in> NF_terms Q" 
      proof
        fix i
        from step[of i]
        have "\<And> v. v \<in> set (args s \<cdot> \<mu> ^^ i |_ p) \<Longrightarrow> v \<in> NF_terms Q" by auto
        with u  show "u \<cdot> \<mu> ^^ i \<in> NF_terms Q" unfolding subt_at_subst[OF p] sp by auto
      qed
    qed (insert p r s t, auto)
  qed
qed

definition check_rstep' :: 
     "('f :: show, 'v :: show) rules \<Rightarrow>
     pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow>
     shows check"
where "check_rstep' \<equiv> check_prop_rstep' (\<lambda> _. OK)"

lemma check_rstep'[simp]:
  "isOK (check_rstep' R p r s t) = (\<exists> \<sigma>. (s, t) \<in> rstep_r_p_s (set R) r p \<sigma>)"
  unfolding check_rstep'_def by (simp add: prop_rstep_rstep)

(* TODO: move *)
lemma case_ok_pair[simp]: "isOK(case lr of (l,r) \<Rightarrow> P l r) = (case lr of (l,r) \<Rightarrow> isOK(P l r))" by auto
lemma case_ok_option[simp]: "isOK(case opt of None \<Rightarrow> P | Some v \<Rightarrow> Q v) = (case opt of None \<Rightarrow> isOK(P) | Some v \<Rightarrow> isOK(Q v))" by (cases opt, auto)        
  

definition check_qrstep ::    "(('f::show, 'v::show) term \<Rightarrow> bool) \<Rightarrow>
     ('f, 'v) rules \<Rightarrow>
     pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow>
     shows check"
where "check_qrstep nf \<equiv> check_prop_rstep (\<lambda> t. check (nf t) (shows t +@+ shows '' is not in Q-normal form''))"



lemma check_qrstep_sound:
  assumes nf: "\<And>t. nf t \<longleftrightarrow> t \<in> NF_terms (set Q)"
    and  ok: "isOK (check_qrstep nf R p r s t)"
  shows "\<exists> \<sigma> r. (s, t) \<in> qrstep_r_p_s (set Q) (set R) r p \<sigma>"
  using check_prop_rstep_sound[OF ok[unfolded check_qrstep_def]]
  unfolding isOK_check prop_rstep_qrstep[symmetric] using nf by auto


lemma check_qrstep_complete:
  assumes nf: "\<And>t. nf t \<longleftrightarrow> t \<in> NF_terms (set Q)"
  and ok: "\<exists> \<sigma>. (s, t) \<in> qrstep_r_p_s (set Q) (set R) r p \<sigma>"
  shows "isOK (check_qrstep nf R p r s t)"
  unfolding check_qrstep_def nf 
  by (rule check_prop_rstep_complete, insert ok[unfolded prop_rstep_qrstep[symmetric]], auto)

lemma check_qrstep_qrstep:
  assumes nf: "\<And>t. nf t \<longleftrightarrow> t \<in> NF_terms (set Q)"
    and  ok: "isOK (check_qrstep nf R p r s t)"
  shows "(s, t) \<in> qrstep (set Q) (set R)"
  unfolding qrstep_qrstep_r_p_s_conv
  using check_qrstep_sound[OF nf ok] by auto

definition check_qrstep_subst :: "(('f::show, 'v::show) term \<Rightarrow> ('f,'v)redex_prob check) \<Rightarrow>
     ('f, 'v) rules \<Rightarrow>
     pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow>
     shows check"
where "check_qrstep_subst cni \<equiv> let main = check_prop_rstep (\<lambda> t. cni t <+? (\<lambda> _. shows t +@+ shows '' mu ^^ i is not in Q-normal form for all i'')) in
   (\<lambda> R p r s t. do { 
      check (is_Fun (fst r)) (shows ''loop check requires lhss to be non-variable'');
      main R p r s t
   })"

lemma check_qrstep_subst_pos: assumes cni: "cni = check_NF_iteration \<mu> Q"
  and ok: "isOK(check_qrstep_subst cni R p r s t)"
  shows "\<exists> \<sigma> r'. (s \<cdot> mk_subst \<mu> ^^ i, t \<cdot> mk_subst \<mu> ^^ i) \<in> qrstep_r_p_s (set Q) (set R) r' p \<sigma> \<and> r =v r'"
proof -
  from ok[unfolded check_qrstep_subst_def Let_def]
  have nvar: "is_Fun (fst r)" by auto
  from check_prop_rstep_sound[OF ok[unfolded check_qrstep_subst_def Let_def cni, simplified, THEN conjunct2], unfolded isOK_update_error check_NF_iteration]
  obtain \<sigma> r' where step: "(s,t) \<in> prop_rstep_r_p_s (\<lambda> t. \<forall> i. t \<cdot> mk_subst \<mu> ^^ i \<in> NF_terms (set Q)) (set R) r' p \<sigma>" 
    and r': "r =v r'" by auto
  from r' nvar have nvar: "is_Fun (fst r')" unfolding eq_rule_mod_vars_def instance_rule_def by (cases "fst r", auto)
  from step show ?thesis unfolding prop_rstep_qrstep_subst[OF nvar] using r' by blast
qed

lemma check_qrstep_subst: assumes cni: "cni = check_NF_iteration \<mu> Q"
  and ok: "isOK(check_qrstep_subst cni R p r s t)"
  shows "(s \<cdot> mk_subst \<mu> ^^ i, t \<cdot> mk_subst \<mu> ^^ i) \<in> qrstep (set Q) (set R)"
  using check_qrstep_subst_pos[OF cni ok] 
  unfolding qrstep_qrstep_r_p_s_conv by blast

definition check_rqrstep_subst :: "(('f::show, 'v::show) term \<Rightarrow> ('f,'v)redex_prob check) \<Rightarrow>
     ('f, 'v) rules \<Rightarrow>
     ('f, 'v) rule \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow>
     shows check"
where "check_rqrstep_subst cni \<equiv> \<lambda> R. check_qrstep_subst cni R \<epsilon>"

lemma check_rqrstep_subst: assumes cni: "cni = check_NF_iteration \<mu> Q"
  and ok: "isOK(check_rqrstep_subst cni R r s t)"
  shows "(s \<cdot> mk_subst \<mu> ^^ i, t \<cdot> mk_subst \<mu> ^^ i) \<in> rqrstep (set Q) (set R)"
  using check_qrstep_subst_pos[OF cni ok[unfolded check_rqrstep_subst_def], of i] 
  unfolding qrstep_r_p_s_def rqrstep_def Let_def by auto

definition
  check_no_defined_root :: "(('f \<times> nat) \<Rightarrow> bool) \<Rightarrow> ('f::show, 'v::show) term \<Rightarrow> shows check"
where
  "check_no_defined_root isdef t =
    check (\<not> isdef (the (root t))) (
      ''the root of '' +#+ shows t +@+ shows_string '' is defined'')"

lemma check_no_defined_root_sound:
  assumes "\<And>fn. isdef fn \<longleftrightarrow> defined (set R) fn"
  shows "isOK (check_no_defined_root isdef t) \<longleftrightarrow> \<not> defined (set R) (the (root t))"
  using assms by (cases t) (auto simp: check_no_defined_root_def)

definition
  check_rqrstep ::
    "(('f::show, 'v::show) term \<Rightarrow> bool) \<Rightarrow> ('f, 'v) rules \<Rightarrow>
     ('f, 'v) rule \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow>
     shows check"
where
  "check_rqrstep nf R rule s t \<equiv> check_qrstep nf R \<epsilon> rule s t"

lemma check_rqrstep_sound':
  assumes nf: "\<And>t. nf t \<longleftrightarrow> t \<in> NF_terms (set Q)"
    and ok: "isOK (check_rqrstep nf R r s t)"
  shows "\<exists>\<sigma> r. (s, t) \<in> qrstep_r_p_s (set Q) (set R) r \<epsilon> \<sigma>"
  using check_qrstep_sound[OF assms[unfolded check_rqrstep_def]] .

lemma check_rqrstep_sound:
  assumes nf: "\<And>t. nf t \<longleftrightarrow> t \<in> NF_terms (set Q)"
    and ok: "isOK (check_rqrstep nf R r s t)"
  shows "(s, t) \<in> rqrstep (set Q) (set R)"
  using check_qrstep_sound[OF assms[unfolded check_rqrstep_def]]
  unfolding rqrstep_def qrstep_r_p_s_def by auto

type_synonym  ('f, 'v) prseq = "(pos \<times> ('f, 'v) rule \<times> bool \<times> ('f, 'v) term) list"
type_synonym  ('f, 'v) rseq = "(pos \<times> ('f, 'v) rule \<times> ('f, 'v) term) list"

fun
  check_qsteps ::
    "(('f::show, 'v::show) term \<Rightarrow> bool) \<Rightarrow>
     ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow>
     ('f, 'v) prseq \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_qsteps nf P R [] s u = check (s = u) (
    ''the last term of the rewrite sequence'' +#+ shows_nl +@+ shows s +@+ shows_nl +@+
    ''does not correspond to the goal term''  +#+ shows_nl +@+ shows u +@+ shows_nl)"
| "check_qsteps nf P R ((_, r, True, t) # prts) s u = do {
    check_rqrstep nf P r s t;
    check_qsteps nf P R prts t u
  }"
| "check_qsteps nf P R ((p, r, False, t) # prts) s u = do {
    check_qrstep nf R p r s t;
    check_qsteps nf P R prts t u
  }"

lemma check_qsteps_sound:
  assumes nf: "\<And>t. nf t \<longleftrightarrow> t \<in> NF_terms (set Q)"
    and ok: "isOK (check_qsteps nf P R rseq s t)"
  shows "(s, t) \<in> (rqrstep (set Q) (set P) \<union> qrstep (set Q) (set R))^^(length rseq)"
proof -
  obtain PR where PR: "PR = rqrstep (set Q) (set P) \<union> qrstep (set Q) (set R)" by auto
  from ok show ?thesis unfolding PR[symmetric] 
  proof (induct rseq arbitrary: s)
    case Nil thus ?case by simp
  next
    case (Cons prt rseq)
    obtain p r or u where prt: "prt = (p, r, or, u)" by (cases prt, blast)
    from Cons[unfolded prt]
    have IH: "(u, t) \<in> PR ^^ (length rseq)" by (cases or, induct rseq) auto
    have "(s, u) \<in> PR"
    proof (cases or)
      case True
      with Cons[unfolded prt]
      have "isOK (check_rqrstep nf P r s u)" by simp
      from check_rqrstep_sound[OF nf this]
        have "(s, u) \<in> rqrstep (set Q) (set P)" .
      thus ?thesis unfolding PR ..
    next
      case False
      with Cons[unfolded prt]
      have "isOK (check_qrstep nf R p r s u)" by simp
      from check_qrstep_qrstep[OF nf this]
        have "(s, u) \<in> qrstep (set Q) (set R)" .
      thus ?thesis unfolding PR ..
    qed
    with IH show ?case unfolding prt o_def using rel_pow_Suc_I2[of s u PR] by simp
  qed
qed

fun
  check_qsteps_subst ::
    "(('f::show, 'v::show) term \<Rightarrow> ('f,'v)redex_prob check) \<Rightarrow>
     ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow>
     ('f, 'v) prseq \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_qsteps_subst cni P R [] s u = check (s = u) (
    ''the last term of the rewrite sequence'' +#+ shows_nl +@+ shows s +@+ shows_nl +@+
    ''does not correspond to the goal term''  +#+ shows_nl +@+ shows u +@+ shows_nl)"
| "check_qsteps_subst cni P R ((_, r, True, t) # prts) s u = do {
    check_rqrstep_subst cni P r s t;
    check_qsteps_subst cni P R prts t u
  }"
| "check_qsteps_subst cni P R ((p, r, False, t) # prts) s u = do {
    check_qrstep_subst cni R p r s t;
    check_qsteps_subst cni P R prts t u
  }"

lemma check_qsteps_subst_sound:
  assumes cni: "cni = check_NF_iteration \<mu> Q"
    and ok: "isOK (check_qsteps_subst cni P R rseq s t)"
  shows "(s \<cdot> mk_subst \<mu> ^^ i, t \<cdot> mk_subst \<mu> ^^ i) \<in> (rqrstep (set Q) (set P) \<union> qrstep (set Q) (set R))^^(length rseq)"
proof -
  obtain PR where PR: "PR = rqrstep (set Q) (set P) \<union> qrstep (set Q) (set R)" by auto
  from ok show ?thesis unfolding PR[symmetric] 
  proof (induct rseq arbitrary: s)
    case Nil thus ?case by simp
  next
    case (Cons prt rseq)
    obtain p r or u where prt: "prt = (p, r, or, u)" by (cases prt, blast)
    let ?\<sigma> = "mk_subst \<mu> ^^ i"
    note Cons = Cons[unfolded prt]
    from Cons(2) have ok: "isOK(check_qsteps_subst cni P R rseq u t)" by (cases or, auto)
    have IH: "(u \<cdot> ?\<sigma> , t \<cdot> ?\<sigma>) \<in> PR ^^ (length rseq)" 
      by (rule Cons(1)[OF ok])
    have "(s \<cdot> ?\<sigma>, u \<cdot> ?\<sigma>) \<in> PR"
    proof (cases or)
      case True
      with Cons[unfolded prt]
      have "isOK (check_rqrstep_subst cni P r s u)" by simp
      from check_rqrstep_subst[OF cni this]
        have "(s \<cdot> ?\<sigma>, u \<cdot> ?\<sigma>) \<in> rqrstep (set Q) (set P)" .
      thus ?thesis unfolding PR ..
    next
      case False
      with Cons[unfolded prt]
      have "isOK (check_qrstep_subst cni R p r s u)" by simp
      from check_qrstep_subst[OF cni this]
        have "(s \<cdot> ?\<sigma>, u \<cdot> ?\<sigma>) \<in> qrstep (set Q) (set R)" .
      thus ?thesis unfolding PR ..
    qed
    with IH show ?case unfolding prt o_def using rel_pow_Suc_I2[of "s \<cdot> ?\<sigma>" "u \<cdot> ?\<sigma>" PR] by simp
  qed
qed

definition
  check_qrsteps ::
    "(('f::show, 'v::show) term \<Rightarrow> bool) \<Rightarrow> ('f, 'v) rules \<Rightarrow>
     ('f, 'v) rseq \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_qrsteps nf R prts s u \<equiv>
    check_qsteps nf [] R (map (\<lambda>(p, r, t). (p, r, False, t)) prts) s u"

lemma check_qrsteps_sound:
  assumes nf: "\<And>t. nf t \<longleftrightarrow> t \<in> NF_terms (set Q)"
    and ok: "isOK (check_qrsteps nf R rseq s t)"
  shows "(s, t) \<in> (qrstep (set Q) (set R))^^(length rseq)"
proof -
  have empty: "rqrstep (set Q) {} \<union> qrstep (set Q) (set R) = qrstep (set Q) (set R)"
    unfolding rqrstep_def qrstep_r_p_s_def by auto
  from check_qsteps_sound[OF assms[unfolded check_qrsteps_def]]
    show ?thesis by (auto simp: Let_def empty)
qed

definition
  check_qrsteps_subst ::
    "(('f::show, 'v::show) term \<Rightarrow> ('f,'v)redex_prob check) \<Rightarrow>
     ('f, 'v) rules \<Rightarrow>
     ('f, 'v) rseq \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_qrsteps_subst cni R prts s u \<equiv>
    check_qsteps_subst cni [] R (map (\<lambda>(p, r, t). (p, r, False, t)) prts) s u"

lemma check_qrsteps_subst_sound:
  assumes cni: "cni = check_NF_iteration \<mu> Q"
    and ok: "isOK (check_qrsteps_subst cni R rseq s t)"
  shows "(s \<cdot> mk_subst \<mu> ^^ i, t \<cdot> mk_subst \<mu> ^^ i) \<in> (qrstep (set Q) (set R))^^(length rseq)"
proof -
  have empty: "rqrstep (set Q) {} \<union> qrstep (set Q) (set R) = qrstep (set Q) (set R)"
    unfolding rqrstep_def qrstep_r_p_s_def by auto
  from check_qsteps_subst_sound[OF assms[unfolded check_qrsteps_subst_def]]
    show ?thesis by (auto simp: Let_def empty)
qed

end
