(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RPO_Impl
imports RPO Multiset_Extension_Impl
begin

(* Implementation of RPO *)

function
  rpo_unbounded_impl :: "('f \<Rightarrow> nat \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> bool \<times> bool) \<times> ('f \<Rightarrow> bool) \<Rightarrow>
  ('f \<Rightarrow> nat \<Rightarrow> order_tag) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool \<times> bool"
where
  "rpo_unbounded_impl _  _ (Var x) (Var y) = (False, x = y)" |
  "rpo_unbounded_impl pr _ (Var x) (Fun g ts) = (False, ts = [] \<and> (snd pr) g)" |
  "rpo_unbounded_impl pr c (Fun f ss) (Var y) = (let con = (\<exists> s \<in> set ss. snd (rpo_unbounded_impl pr c s (Var y))) in (con,con))" |
  "rpo_unbounded_impl pr c (Fun f ss) (Fun g ts) = (
    if (\<exists> s \<in> set ss. snd (rpo_unbounded_impl pr c s (Fun g ts)))
    then (True,True)
    else (let (prs,prns) = (fst pr) f (length ss) g (length ts) in 
         if prns \<and> (\<forall> t \<in> set ts. fst (rpo_unbounded_impl pr c (Fun f ss) t))
         then if prs
              then (True,True) 
              else let cf = c f (length ss) in
                   if cf = c g (length ts)
                   then if cf = Mul
                        then mul_ext (rpo_unbounded_impl pr c) ss ts
                        else lex_ext_unbounded (rpo_unbounded_impl pr c) ss ts
                   else (ss \<noteq> [] \<and> ts = [], ts = [])
         else (False,False)))"
by pat_completeness auto

termination rpo_unbounded_impl
proof (relation "measures [\<lambda> (pr,c,s,t). size s + size t]")
{
  fix "pr" c and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and x xa y xb x'
  assume s: "xb \<in> set ss" and t: "x' \<in> set ts"
  from s have s1: "size xb \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size x' \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 s t show "((pr, c, xb, x'), pr, c, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,c,s,t). size s + size t]"
    by auto
}
moreover
{
  fix "pr" c and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and x xa xb xc i
  assume "i < length ss" and "i < length ts"
  then obtain s t where si: "s = ss ! i" and s: "s \<in> set ss" and ti: "t = ts ! i" and t: "t \<in> set ts" by auto
  from s have s1: "size s \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size t \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 si ti show "((pr, c, ss ! i, ts ! i), pr, c, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,c,s,t). size s + size t]"
    by auto
}
qed (auto simp: size_simps)

(* Soundness lemma *)
lemma rpo_unbounded_impl_sound:
  shows "rpo_unbounded pr c s t = rpo_unbounded_impl pr c s t" (is "?e pr c s t")
proof (induct rule: rpo_unbounded.induct[where P = ?e])
case (1 p c x y) show ?case by simp
next
case (2 p c x g ts) show ?case by (simp add: prl_nat_def)
next
case (3 p c f ss y) thus ?case by (simp add: Let_def list_ex_iff)
next
case (4 p c f ss g ts) note IH = this
  from 4(1) have s: "\<And> s. s \<in> set ss \<Longrightarrow> ?e p c s (Fun g ts)" by simp
  let ?cond = "\<exists> s \<in> set ss. snd (rpo_unbounded_impl p c s (Fun g ts))"
  show ?case
  proof (cases ?cond)
  case True
    with s have "\<exists> s \<in> set ss. snd (rpo_unbounded p c s (Fun g ts))" by simp
    with True show ?thesis by simp
  next
  case False note oFalse = this
    with s have oFalse2: "\<not> (\<exists>s \<in> set ss. snd (rpo_unbounded p c s (Fun g ts)))"
     by simp
    obtain prns prs where Hsns: "(fst p) f (length ss) g (length ts) = (prs, prns)"
     by (cases "(fst p) f (length ss) g (length ts)") auto
    from 4(2)[OF oFalse2 _ Hsns[symmetric]]
     have t: "\<And> t. t \<in> set ts \<Longrightarrow> ?e p c (Fun f ss) t" by force
    show ?thesis
    proof (cases "prns \<and> (\<forall>t\<in>set ts. fst (rpo_unbounded_impl p c (Fun f ss) t))")
    case False
      show ?thesis
      proof (cases "prns")
      case False thus ?thesis by (simp add: oFalse oFalse2 Hsns)
      next
      case True
        with False have Hf1: "\<not> (\<forall>t\<in>set ts. fst (rpo_unbounded_impl p c (Fun f ss) t))" by simp
        with t have Hf2: "\<not> (\<forall>t\<in>set ts. fst (rpo_unbounded p c (Fun f ss) t))" by auto
        show ?thesis by (simp add: oFalse oFalse2 Hf1 Hf2)
      qed         
    next
    case True
      hence prns: "prns" and Hts: "\<forall>t\<in>set ts. fst (rpo_unbounded_impl p c (Fun f ss) t)"
       by auto
      from Hts and t have Hts2: "\<forall>t\<in>set ts. fst (rpo_unbounded p c (Fun f ss) t)" by auto
      show ?thesis
      proof (cases "prs")
      case True thus ?thesis by (simp add: oFalse oFalse2 Hsns prns Hts Hts2)
      next
      case False note prs = this
        show ?thesis
        proof (cases "c f (length ss) = c g (length ts)")
        case False thus ?thesis
         by (cases "c f (length ss)", simp_all add: oFalse oFalse2 Hsns prns Hts Hts2)
        next
        case True note cfg = this
          show ?thesis
          proof (cases "c f (length ss)")
          case Mul note cf = this
            with cfg have cg: "c g (length ts) = Mul" by simp
            {
              fix x y
              assume x_in_ss: "x \<in> set ss" and y_in_ts: "y \<in> set ts"
              from 4(3)[OF oFalse2 _ Hsns[symmetric] _ prs cfg cf
                x_in_ss y_in_ts] prns Hts2
              have "rpo_unbounded p c x y = rpo_unbounded_impl p c x y" by auto
            }
            note Hrpo1 = this
            have "mul_ext (rpo_unbounded p c) ss ts = mul_ext (rpo_unbounded_impl p c) ss ts"
              by (rule mul_ext_cong[OF refl refl], insert Hrpo1, auto)
            thus ?thesis by (simp add: oFalse oFalse2 Hsns prns prs Hts Hts2 cf cg)
          next
            case Lex note cf = this
            with cfg have cg: "c g (length ts) = Lex" by simp
            from cfg cf have ncf: "c f (length ss) \<noteq> Mul" by simp
            {
              fix i
              assume iss: "i < length ss" and its: "i < length ts"
              from nth_mem_multiset_of[OF iss] and in_multiset_in_set
              have in_ss: "ss ! i \<in> set ss" by force
              from nth_mem_multiset_of[OF its] and in_multiset_in_set
              have in_ts: "ts ! i \<in> set ts" by force
              from 4(4)[OF oFalse2 _ Hsns[symmetric] _ prs cfg ncf iss its]
                prns Hts2
              have "rpo_unbounded p c (ss ! i) (ts ! i)
                = rpo_unbounded_impl p c (ss ! i) (ts ! i)" by simp
            }
            with lex_ext_unbounded_cong[of ss ss ts ts]
            have Hlex1: "lex_ext_unbounded (rpo_unbounded_impl p c) ss ts
              = lex_ext_unbounded (rpo_unbounded p c) ss ts" by metis
            thus ?thesis
              by (simp add: oFalse oFalse2 Hsns prns prs Hts Hts2 cfg cg)
          qed
        qed
      qed
    qed
  qed
qed

(** Extracting code **)

lemma [code]: "rpo_unbounded = rpo_unbounded_impl"
  by (intro ext, simp add: rpo_unbounded_impl_sound)

end