(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Reduction_Pair_Impl
imports
  Reduction_Pair
  QDP_Framework_Impl
  "Q-Restricted_Rewriting_Impl"
  Term_Order_Impl
begin

definition
  mono_redpair_proc ::
    "('dpp, 'f, 'v) dpp_ops \<Rightarrow> ('f::show, 'v::show) redtriple \<Rightarrow> 
     ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> 'dpp proc"
where
  "mono_redpair_proc I rp Premove Rremove dpp = check_return (do {
     let (ps,pns) = dpp_ops.split_pairs I dpp Premove;
     let (rs,rns) = dpp_ops.split_rules I dpp Rremove;
     redtriple.valid rp;
     redtriple.mono rp
       <+? (\<lambda>s. ''problem with monotonicity of strict order'' +#+ shows_nl +@+ s);
     check_allm (redtriple.ns rp) rns
       <+? (\<lambda>s. ''problem when orienting TRS'' +#+ shows_nl +@+ s);
     check_allm (redtriple.s rp) rs
       <+? (\<lambda>s. ''problem when orienting TRS'' +#+ shows_nl +@+ s);
     check_allm (redtriple.ns rp) pns
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s);
     check_allm (redtriple.s rp) ps
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
   } <+? (\<lambda>s. ''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
     (redtriple.desc rp) +@+ shows_nl +@+ s))
   (dpp_spec.delete_pairs_rules I dpp Premove Rremove)"

definition
  rule_removal_tt ::
    "('tp, 'f, 'v) tp_ops \<Rightarrow> ('f::show, 'v::show) redtriple \<Rightarrow> 
    ('f, 'v) rules \<Rightarrow> 'tp proc"
where
  "rule_removal_tt I rp Rremove trs = check_return (do {
     let (rs,rns) = tp_ops.split_rules I trs Rremove;
     redtriple.valid rp;
     redtriple.mono rp
       <+? (\<lambda>s. ''problem with monotonicity of strict order'' +#+ shows_nl +@+ s);
     check_allm (redtriple.ns rp) rns
       <+? (\<lambda>s. ''problem when orienting TRS'' +#+ shows_nl +@+ s);
     check_allm (redtriple.s rp) rs
       <+? (\<lambda>s. ''problem when orienting TRS'' +#+ shows_nl +@+ s)
   } <+? (\<lambda>s. ''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
     (redtriple.desc rp) +@+ shows_nl +@+ s))
   (tp_spec.delete_rules I trs Rremove)"


context dpp_spec
begin

lemma mono_redpair_proc:
  assumes grp: "generic_redtriple_impl grp"
  shows "sound_proc_impl (mono_redpair_proc I (grp rp) ps rs)"
proof 
  fix d d'
  assume fin: "finite_dpp (dpp d')"
  and ok: "mono_redpair_proc I (grp rp)  ps rs d = return d'"
  let ?rp = "grp rp"
  let ?Ps = "set ps"
  let ?Rs = "set rs"
  let ?P = "set (P d)"
  let ?Pw = "set (Pw d)"
  let ?Q = "set (Q d)"
  let ?R = "set (R d)"
  let ?Rw = "set (Rw d)"
  interpret generic_redtriple_impl grp by (rule grp)
  show "finite_dpp (dpp d)"
  proof (cases "split_pairs d ps")
    case (Pair Ps Pr) note pair = this
    from split_pairs_sound[OF this] have Ps: "set Ps = (?P \<union> ?Pw) \<inter> ?Ps" 
      and Pr: "set Pr = (?P \<union> ?Pw) - ?Ps" by (auto simp: dpp_spec_sound)
    show ?thesis 
    proof (cases "split_rules d rs")
      case (Pair Rs Rr) 
      from split_rules_sound[OF this] have Rs: "set Rs = (?R \<union> ?Rw) \<inter> ?Rs" 
        and Rr: "set Rr = (?R \<union> ?Rw) - ?Rs" by (auto simp: dpp_spec_sound)
      note ok = ok[unfolded mono_redpair_proc_def, simplified, simplified Pair pair split]
      from ok have valid: "isOK(redtriple.valid ?rp)" 
        and mono: "isOK(redtriple.mono ?rp)"
        and NS: "isOK (check_allm (redtriple.ns ?rp) (Pr @ Rr))"
        and S: "isOK(check_allm (redtriple.s ?rp) (Ps @ Rs))" 
        and d': "d' = delete_pairs_rules d ps rs" by auto
      from generate_redtriple[OF valid S NS, of Nil]  mono
      obtain S NS NST where "ce_af_redtriple_order S NS NST (redtriple.af ?rp)" and S: "(?P \<union> ?Pw) \<inter> ?Ps \<subseteq> S" "(?R \<union> ?Rw) \<inter> ?Rs \<subseteq> S" and NS: "?P \<union> ?Pw - ?Ps \<subseteq> NS" "?R \<union> ?Rw - ?Rs \<subseteq> NS" and "mono_ce_af_redtriple_order S NS NST (redtriple.af ?rp)" and mono: "one_step_ctxt_closed S" by  (auto simp: dpp_spec_sound Pr Ps Rr Rs)
      have proc: "Reduction_Pair.mono_redpair_proc S NS ?Ps ?Rs (?P,?Pw,?Q,?R,?Rw) (?P-?Ps,?Pw-?Ps,?Q,?R-?Rs,?Rw-?Rs)" using NS S mono by auto
      interpret mono_ce_af_redtriple_order S NS NST "redtriple.af ?rp" by fact
      note fin =  fin[unfolded d' delete_simps]
      show ?thesis unfolding d' dpp_sound
        by (rule ci_proc[OF fin proc mono_redpair_proc_ci])
    qed
  qed
qed
end

context tp_spec
begin

lemma rule_removal_tt:
  assumes grp: "generic_redtriple_impl grp"
  shows "sound_tt_impl (rule_removal_tt I (grp rp) rs)"
proof 
  fix trs trs'
  assume fin: "SN_qrel (qreltrs trs')"
  and ok: "rule_removal_tt I (grp rp) rs trs = return trs'"
  let ?rp = "grp rp"
  let ?Rs = "set rs"
  let ?Q = "set (Q trs)"
  let ?R = "set (R trs)"
  let ?Rw = "set (Rw trs)"
  interpret generic_redtriple_impl grp by (rule grp)
  show "SN_qrel (qreltrs trs)"
  proof (cases "split_rules trs rs")
    case (Pair Rs Rr) 
    from split_rules_sound[OF this] have Rs: "set Rs = (?R \<union> ?Rw) \<inter> ?Rs" 
      and Rr: "set Rr = (?R \<union> ?Rw) - ?Rs" by (auto simp: tp_spec_sound)
    note ok = ok[unfolded rule_removal_tt_def,simplified, simplified Pair]
    from ok have valid: "isOK (redtriple.valid ?rp)"
      and mono: "isOK(redtriple.mono ?rp)"
      and NS: "isOK (check_allm (redtriple.ns ?rp) Rr)"
      and S: "isOK(check_allm (redtriple.s ?rp) Rs)" 
      and trs': "trs' = delete_rules trs rs" by auto
    from generate_redtriple[OF valid S NS, of Nil]  mono
    obtain S NS NST where "ce_af_redtriple_order S NS NST (redtriple.af ?rp)" and S: "(?R \<union> ?Rw) \<inter> ?Rs \<subseteq> S" and NS: "?R \<union> ?Rw - ?Rs \<subseteq> NS" and "mono_ce_af_redtriple_order S NS NST (redtriple.af ?rp)" and mono: "one_step_ctxt_closed S" by  (auto simp: tp_spec_sound Rr Rs)
    have proc: "mono_redpair_tt S NS ?Rs (?Q,?R,?Rw) (?Q,?R-?Rs,?Rw-?Rs)" using NS S mono by auto
    interpret mono_ce_af_redtriple_order S NS NST "redtriple.af ?rp" by fact
    note fin =  fin[unfolded trs' delete_R_Rw_sound]
    show ?thesis unfolding trs' qreltrs_sound
      by (rule rel_tt[OF mono_redpair_tt _ fin], rule proc)
  qed
qed
end

end
