(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Term_Impl
imports
  Term
  Monadic_Check
  Show
begin

(* TODO: why cannot we use substpow.simps directly *)
primrec substpow_impl :: "('f, 'v) subst \<Rightarrow> nat \<Rightarrow> ('f, 'v) subst" where
  "substpow_impl \<sigma> 0  = Var"
| "substpow_impl \<sigma> (Suc n) = \<sigma> \<circ>s substpow_impl \<sigma> n"

lemma substpow_impl[simp]: "substpow_impl = (\<lambda> \<sigma> n . \<sigma> ^^ n)"
proof (rule ext, rule ext)
  fix \<sigma> n
  show "substpow_impl \<sigma> n = \<sigma> ^^ n"
    by (induct n, auto)
qed

subsection {*@{text "show"} Instance for Terms.*}

(*shows method for terms that is parametrized over shows methods for function
symbols and variables*)
fun shows_term' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> ('f, 'v) term \<Rightarrow> shows" where
  "shows_term' fun var (Var x) = var x"
| "shows_term' fun var (Fun f ts) = (
    fun f +@+ shows_list_gen id [] ''('' '', '' '')'' (map (shows_term' fun var) ts)
  )"

instantiation "term" :: ("show", "show") "show" begin
  abbreviation "shows_term \<equiv> shows_term' shows shows"
  definition "shows_prec (d::nat) = shows_term"
  definition "shows_list_term (is::('f::show, 'v::show) term list) = shows_list_aux shows is"

  lemma assoc_term:
    "shows_prec d (t::('f::show, 'v::show) term) r @ s = shows_prec d t (r @ s)"
  proof (induct t arbitrary: r s)
    case (Var x) show ?case unfolding shows_prec_term_def by simp
  next
    case (Fun f ts)
    hence "\<And>r s. \<forall>t\<in>set ts. id (shows_prec d t) r @ s = id (shows_prec d t) (r @ s)" by auto
    hence elt: "\<And>r s. \<forall>t\<in>set (map shows_term ts). id t r @ s = id t (r @ s)"
      unfolding shows_prec_term_def by auto
    show ?case
      unfolding shows_prec_term_def
      unfolding shows_term'.simps o_def
      unfolding assoc
      unfolding shows_list_gen_assoc[OF elt,of "[]" "''(''" "'', ''" "'')''" r s]
      using Fun by simp
  qed
  lemma assoc_term_elt:
    "\<forall>(t::('f::show, 'v::show) term)\<in>set ts. shows t r @ s = shows t (r @ s)"
    using assoc_term by auto
 instance by
    (intro_classes,rule assoc_term,unfold shows_list_term_def)
    (rule shows_list_aux_assoc[OF assoc_term_elt])
end


subsection {* @{text "shows"} Instance for Contexts *}

fun shows_ctxt' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> shows" where
  "shows_ctxt' fun var (Hole) = shows_string ''[]''"
| "shows_ctxt' fun var (More f ss1 D ss2) = (
    fun f +@+ ''('' +#+
    shows_list_gen (shows_term' fun var) [] [] '', '' '', '' ss1 +@+
    shows_ctxt' fun var D +@+
    shows_list_gen (shows_term' fun var) '')'' '', '' '', '' '')'' ss2
  )"

instantiation ctxt :: ("show", "show") "show" begin
  abbreviation "shows_ctxt \<equiv> shows_ctxt' shows shows"
  definition "shows_prec (d::nat) = shows_ctxt"
  definition "shows_list_ctxt (cs::('f::show,'v::show) ctxt list) \<equiv> shows_list_aux shows cs"

  lemma assoc_ctxt:
    "shows_prec d (c::('f::show, 'v::show) ctxt) r @ s = shows_prec d c (r @ s)"
    by (induct c arbitrary: r s)
       (auto simp: shows_prec_ctxt_def shows_prec_term_def[symmetric, of 0])

  lemma assoc_ctxt_elt:
    "\<forall>(c::('f::show, 'v::show) ctxt)\<in>set cs. shows_prec d c r @ s = shows_prec d c (r @ s)"
  using assoc_ctxt by auto

  instance by
    (intro_classes,rule assoc_ctxt,unfold shows_list_ctxt_def)
    (rule shows_list_aux_assoc[OF assoc_ctxt_elt])
end

fun
  poss_list :: "('f, 'v) term \<Rightarrow> pos list"
where
  "poss_list (Var x) = [\<epsilon>]" |
  "poss_list (Fun f ss) = (\<epsilon> # concat (map (\<lambda> (i,ps).
    map ( op <# i) ps) (zip [0 ..< length ss] (map poss_list ss))))"

lemma poss_list_sound[simp]: "set (poss_list s) = poss s" 
proof (induct s, simp)
  case (Fun f ss)
  let ?z = "zip [0..<length ss] (map poss_list ss)"
  have len: "length ?z = length ss" by auto
  have "(\<Union>a\<in>set ?z.
         set (split (\<lambda>i. map (op <# i)) a)) = 
       {i <# p |i p. i < length ss \<and> p \<in> poss (ss ! i)}" (is "?l = ?r")
  proof (rule set_eqI)
    fix ip
    show "(ip \<in> ?l) = (ip \<in> ?r)"
    proof
      assume "ip \<in> ?l" 
      from this obtain ipI where 
	z: "ipI \<in> set ?z" and 
	ip: "ip \<in> set (split (\<lambda> i. map (op <# i)) ipI)" 
	by auto     
      from z obtain i where i: "i < length ?z" and zi: "?z ! i = ipI"
        by (force simp: set_conv_nth)      
      with ip Fun show "ip \<in> ?r" by auto
    next
      assume "ip \<in> ?r"
      from this obtain i p where i: "i < length ss" and "p \<in> poss (ss ! i)"
	and ip: "ip = i <# p" by auto
      with Fun have p: "p \<in> set (poss_list (ss ! i))" and iz: "i < length ?z" by auto
      from i have id: "?z ! i = (i, poss_list (ss ! i))" (is "_ = ?ipI") by auto
      from iz have  "?z ! i \<in> set ?z" by (rule nth_mem)
      with id have inZ: "?ipI \<in> set ?z" by auto
      from p have "i <# p \<in> set (split (\<lambda> i. map (op <# i)) ?ipI)" by auto
      with inZ ip show "ip \<in> ?l" by force
    qed
  qed
  with Fun show ?case by simp
qed

declare poss_list.simps[simp del]

fun in_poss :: "pos \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "in_poss \<epsilon> _ = True"
      | "in_poss (PCons i p) (Fun f ts) = (i < length ts \<and> in_poss p (ts ! i))"
      | "in_poss (PCons i p) (Var _) = False"

lemma in_poss[simp]: "in_poss p t = (p \<in> poss t)"
  by (induct rule: in_poss.induct, auto)

declare in_poss.simps[simp del]


text {*Resolve problems with recurisve dependencies in generated code.*}
lemma [code]:
  shows "HOL.eq (Var x) (Var y) \<longleftrightarrow> HOL.eq x y"
  and "HOL.eq (Fun f ts) (Fun g us) \<longleftrightarrow> HOL.eq f g \<and> list_all2 HOL.eq ts us"
  by (simp_all add: list_all2_eq [symmetric])

subsection {* Useful abstractions *}

text {* Given that we perform the same operations on terms in order to get
a list of the variables and a list of the functions, we define functions
that run through the term and perform these actions. *}

fun contains_var_term :: "'v \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" where
  "contains_var_term x (Var y) = (x = y)"
| "contains_var_term x (Fun _ ts) = list_ex (contains_var_term x) ts"

lemma contains_var_term_sound[simp]:
  "contains_var_term x t \<longleftrightarrow> x \<in> vars_term t"
  by (induct t) (auto simp: Bex_set_list_ex[symmetric])

(*use efficient implementation for code-generation*)
lemma [code_unfold]: "x \<in> vars_term t = contains_var_term x t" by simp

fun vars_term_list' :: "('f, 'v) term \<Rightarrow> 'v list" where
  "vars_term_list' (Var x) = [x]"
| "vars_term_list' (Fun _ ts) = concat (map vars_term_list' ts)"

lemma vars_term_list'_sound[simp]: "set (vars_term_list' t) = vars_term t"
  by (induct t) simp_all
declare vars_term_list'.simps[simp del]

definition vars_term_list :: "('f, 'v) term \<Rightarrow> 'v list" where
  "vars_term_list = remdups \<circ> vars_term_list'"

lemma vars_term_list_sound[simp]:
  "set (vars_term_list t) = vars_term t"
  by (simp add: vars_term_list_def)

text {*The list of function symbols in a term (without removing duplicates).*}

fun funs_term_list' :: "('f, 'v) term \<Rightarrow> 'f list" where
  "funs_term_list' (Var _) = []"
| "funs_term_list' (Fun f ts) = f # concat (map funs_term_list' ts)"

lemma funs_term_list'_sound[simp]:
  "set (funs_term_list' t) = funs_term t"
  by (induct t) simp_all
declare funs_term_list'.simps[simp del]

definition funs_term_list :: "('f, 'v) term \<Rightarrow> 'f list" where
  "funs_term_list \<equiv> remdups \<circ> funs_term_list'"

lemma funs_term_list_sound[simp]:
  "set (funs_term_list t) = funs_term t"
  by (simp add: funs_term_list_def)

abbreviation funs_terms_list where "funs_terms_list Q \<equiv> remdups (concat (map funs_term_list Q))"


text {*The list of function symbol and arity pairs in a term
(without removing duplicates).*}

fun funas_term_list' :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) list" where
  "funas_term_list' (Var _) = []"
| "funas_term_list' (Fun f ts) = (f, length ts) # concat (map funas_term_list' ts)"

lemma funas_term_list'_sound[simp]:
  "set (funas_term_list' t) = funas_term t"
  by (induct t) simp_all
declare funas_term_list'.simps[simp del]

definition funas_term_list :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) list" where
  "funas_term_list \<equiv> remdups \<circ> funas_term_list'"

lemma funas_term_list_sound[simp]:
  "set (funas_term_list t) = funas_term t"
  by (simp add: funas_term_list_def)

definition funas_args_term_list' :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) list" where
  "funas_args_term_list' t = concat (map funas_term_list' (args t))"

lemma funas_args_term_list'_sound[simp]:
  "set (funas_args_term_list' t) = funas_args_term t"
  unfolding funas_args_term_def funas_args_term_list'_def by simp

definition funas_args_term_list :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) list" where
  "funas_args_term_list = remdups \<circ> funas_args_term_list'"

lemma funas_args_term_list_sound[simp]:
  "set (funas_args_term_list t) = funas_args_term t"
  by (simp add: funas_args_term_list_def)

lemma vars_term_list'_map_term: "vars_term_list' (map_term fg t) = vars_term_list' t"
proof (induct t)
  case (Var x) thus ?case by (simp add: vars_term_list'.simps)
next
  case (Fun f ss)
  show ?case by (simp add: vars_term_list'.simps o_def, insert Fun, induct ss, auto)
qed

lemma funs_term_list'_map_term:
  "funs_term_list' (map_term fg t) = map fg (funs_term_list' t)"
proof (induct t)
  case (Var x) show ?case by (simp add: funs_term_list'.simps)
next
  case (Fun f ts)
  show ?case 
    by (simp add: funs_term_list'.simps, insert Fun, induct ts, auto)
qed

lemma linear_vars_term_list': assumes "linear_term t"
  shows "length (filter (op = x) (vars_term_list' t)) \<le> 1"
  using assms
proof (induct t)
  case (Var y)
  show ?case by (auto simp: vars_term_list'.simps)
next
  case (Fun f ss)
  show ?case 
  proof (rule ccontr)
    assume "\<not> ?thesis"
    from this[unfolded vars_term_list'.simps]
    have len: "2 \<le> length (filter (op = x) (concat (map vars_term_list' ss)))" (is "_ \<le> length ?xs") by auto
    from len obtain y ys where xs: "?xs = y # ys" by (cases ?xs, auto)
    from len[unfolded xs] obtain z zs where ys: "ys = z # zs" by (cases ys, auto)
    from xs[unfolded ys] have "{y,z} \<subseteq> set ?xs" by auto
    from this[unfolded set_filter] have "y = x" and "z = x" by auto
    from xs[unfolded ys this] have xs: "?xs = x # x # zs" by auto
    {
      fix s
      assume s: "s \<in> set ss"
      with Fun(2)[unfolded linear_term.simps list_all_iff]
      have "linear_term s" by auto
      note Fun(1)[OF s this]
    }
    from this Fun(2)[unfolded linear_term.simps list_all_iff] xs
    show False
    proof (induct ss)
      case Nil thus ?case by simp
    next
      case (Cons s ss) note oCons = this
      from Cons(3) have part: "is_partition (map vars_term ss) \<and> (\<forall> s \<in> set ss. linear_term s)" and lin: "linear_term s"  using is_partition_Cons by auto
      from Cons(1)[OF _ part] Cons(2) 
      have ind: "filter (op = x) (concat (map vars_term_list' ss)) \<noteq> x # x # zs" by auto
      show ?case 
      proof (cases "filter (op = x) (vars_term_list' s)")
        case Nil
        with Cons(4) ind show False by auto
      next
        case (Cons y ys)
        let ?s = "filter (op = x) (vars_term_list' s)"
        let ?ss = "filter (op = x) (concat (map vars_term_list' ss))"
        from Cons oCons(4) have "?s = x # ys" by auto
        with oCons(2)[of s] have sx: "?s = [x]"
          by auto
        with oCons(4) have ssx: "?ss = x # zs" by auto
        from sx have "x \<in> set ?s" by auto
        from this[unfolded set_filter vars_term_list'_sound]
        have sx: "x \<in> vars_term s" by auto 
        from ssx have "x \<in> set ?ss" by auto
        from this[unfolded set_filter vars_term_list'_sound]
        obtain t where tx: "x \<in> vars_term t" and t: "t \<in> set ss" by auto
        from t[unfolded set_conv_nth] obtain i where i: "i < length ss" and
          t: "t = ss ! i" by auto
        from oCons(3)[THEN conjunct1, unfolded is_partition_def, THEN spec[of _ "Suc i"], THEN mp, THEN spec[of _ 0]] i tx[unfolded t] sx
        show False by auto
      qed
    qed  
  qed
qed


lemma vars_term_list_map_term:
  "vars_term_list (map_term fg t) = vars_term_list t"
  unfolding vars_term_list_def vars_term_list'_map_term o_def ..

fun supteq_list :: "('f, 'v) term \<Rightarrow> ('f, 'v) term list" where
  "supteq_list (Var x) = [Var x]"
| "supteq_list (Fun f ts) = Fun f ts # concat (map supteq_list ts)"

fun supt_list :: "('f, 'v) term \<Rightarrow> ('f, 'v) term list" where
  "supt_list (Var x) = []"
| "supt_list (Fun f ts) = concat (map supteq_list ts)"

lemma supteq_list[simp]: "set (supteq_list t) = {s. t \<unrhd> s}"
proof (rule set_eqI, simp)
  fix s
  show "s \<in> set(supteq_list t) = (t \<unrhd> s)"
  proof (induct t, simp add: supteq_var_imp_eq)
    case (Fun f ss)
    show ?case
    proof (cases "Fun f ss = s", (auto)[1])
      case False
      show ?thesis
      proof
        assume "Fun f ss \<unrhd> s"
        with False have sup: "Fun f ss \<rhd> s" using supteq_supt_conv by auto
        obtain C where "C \<noteq> \<box>" and "Fun f ss = C\<langle>s\<rangle>" using sup by auto
        then obtain b D a where "Fun f ss = Fun f (b @ D\<langle>s\<rangle> # a)" by (cases C, auto)
        hence D: "D\<langle>s\<rangle> \<in> set ss" by auto
        with Fun[OF D] ctxt_imp_supteq[of D s] obtain t where "t \<in> set ss" and "s \<in> set (supteq_list t)" by auto
        thus "s \<in> set (supteq_list (Fun f ss))" by auto
      next
        assume "s \<in> set (supteq_list (Fun f ss))"
        with False obtain t where t: "t \<in> set ss" and "s \<in> set (supteq_list t)" by auto
        with Fun[OF t] have "t \<unrhd> s" by auto
        with t show "Fun f ss \<unrhd> s" by auto
      qed
    qed
  qed
qed

lemma supt_list_sound[simp]: "set (supt_list t) = {s. t \<rhd> s}"
  by (cases t) auto

fun supt_impl :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" where
  "supt_impl (Var x) t = False"
| "supt_impl (Fun f ss) t = (t \<in> set ss \<or> (list_ex (\<lambda>s. supt_impl s t) ss))"

lemma supt_impl[code]: "op \<rhd> = {(x, y). supt_impl x y}"
proof (intro equalityI subsetI2)
  fix s t assume "s \<rhd> t" hence "supt_impl s t"
  proof (induct s)
    case (Var x) thus ?case by auto
  next
    case (Fun f ss) thus ?case
    proof (cases "t \<in> set ss")
      case True thus ?thesis by (simp)
    next
      case False
      assume "\<And>s. \<lbrakk>s \<in> set ss; s \<rhd> t\<rbrakk> \<Longrightarrow> supt_impl s t"
        and "Fun f ss \<rhd> t" and "t \<notin> set ss"
      moreover from `Fun f ss \<rhd> t` obtain s where "s \<in> set ss" and "s \<rhd> t"
        by (cases rule: supt.cases) (simp_all add: `t \<notin> set ss`)
      ultimately have "supt_impl s t" by simp
      with `s \<in> set ss` show ?thesis by (auto simp: list_ex_iff)
    qed
  qed
  thus "(s, t) \<in> {(x, y). supt_impl x y}" by (simp add: split_def)
next
  fix s t assume "(s, t) \<in> {(x, y). supt_impl x y}"
  hence "supt_impl s t" by (simp)
  thus "s \<rhd> t"
  proof (induct s)
    case (Var x) thus ?case by simp
  next
    case (Fun f ss)
    hence "t \<in> set ss \<or> (\<exists>s\<in>set ss. supt_impl s t)" by (simp add: list_ex_iff)
    thus ?case
    proof
      assume "t \<in> set ss" thus ?case by (auto intro: supt.intros)
    next
      assume "\<exists>s\<in>set ss. supt_impl s t"
      then obtain s where "s \<in> set ss" and "supt_impl s t" by auto
      with Fun have "s \<rhd> t" by auto
      with `s \<in> set ss` show ?thesis by (auto intro: supt.intros)
    qed
  qed
qed

lemma supteq_impl[code]: "op \<unrhd> = {(x, y). x = y \<or> supt_impl x y}"
  unfolding supteq_supt_set_conv
  by (auto simp: supt_impl)

fun
  linear_term_impl :: "'v list \<Rightarrow> ('f, 'v) term \<Rightarrow> ('v list) option"
where
  "linear_term_impl xs (Var x) = (if x \<in> set xs then None else Some (x # xs))" |
  "linear_term_impl xs (Fun _ []) = Some xs" |
  "linear_term_impl xs (Fun f (t # ts)) = (case linear_term_impl xs t of 
    None \<Rightarrow> None
  | Some ys \<Rightarrow> linear_term_impl ys (Fun f ts))"

lemma [code]: "linear_term t = (case linear_term_impl [] t of None \<Rightarrow> False | Some _ \<Rightarrow> True)" 
proof -
  {
    fix xs ys
    have "(linear_term_impl xs t = None \<longrightarrow> (set xs \<inter> vars_term t \<noteq> {} \<or> \<not> linear_term t)) \<and> 
      (linear_term_impl xs t = Some ys \<longrightarrow> (set ys = set xs \<union> vars_term t) \<and> set xs \<inter> vars_term t = {} \<and> linear_term t)"
    proof (induct arbitrary: ys rule: linear_term_impl.induct, force, simp add: is_partition_Nil)
      case (3 xs f t ts zs)
      show ?case 
      proof (cases "linear_term_impl xs t")
        case None
        with 3 show ?thesis by auto
      next
        case (Some ys)
        note some = this
        with 3 have rec1: "set ys = set xs \<union> vars_term t \<and> set xs \<inter> vars_term t = {} \<and> linear_term t" by auto
        show ?thesis 
        proof (cases "linear_term_impl ys (Fun f ts)")
          case None 
          with rec1 Some have res: "linear_term_impl xs (Fun f (t # ts)) = None" by simp
          from None 3(2) Some have rec2: "set ys \<inter> vars_term (Fun f ts) \<noteq> {} \<or> \<not> linear_term (Fun f ts)" by simp
          hence "set xs \<inter> vars_term (Fun f (t # ts)) \<noteq> {} \<or> \<not> linear_term (Fun f (t # ts))" 
          proof 
            assume "set ys \<inter> vars_term (Fun f ts) \<noteq> {}" 
            then obtain x where x1: "x \<in> set ys" and x2: "x \<in> vars_term (Fun f ts)" by auto
            show ?thesis 
            proof (cases "x \<in> set xs")
              case True
              with x2 show ?thesis by auto
            next
              case False
              with x1 rec1 have "x \<in> vars_term t" by auto
              with x2 have "\<not> linear_term (Fun f (t # ts))" by (auto simp: is_partition_Cons)
              thus ?thesis ..
            qed
          next
            assume "\<not> linear_term (Fun f ts)" hence "\<not> linear_term (Fun f (t # ts))" by (auto simp: is_partition_Cons)
            thus ?thesis ..
          qed            
          with res show ?thesis by auto
        next
          case (Some us)
          with some have res: "linear_term_impl xs (Fun f (t # ts)) = Some us" by auto          
          {
            assume us: "us = zs"
            from Some[simplified us] 3(2) some 
            have rec2: "set zs = set ys \<union> vars_term (Fun f ts) \<and> set ys \<inter> vars_term (Fun f ts) = {} \<and> linear_term (Fun f ts)" by auto
            from rec1 rec2  
            have part1: "set zs = set xs \<union> vars_term (Fun f (t # ts)) \<and> set xs \<inter> vars_term (Fun f (t # ts)) = {}" (is ?part1) by auto
            from rec1 rec2 have "vars_term t \<inter> vars_term (Fun f ts) = {}" and "linear_term t" and "linear_term (Fun f ts)" by auto
            hence "linear_term (Fun f (t # ts))" (is ?part2) by (auto simp: is_partition_Cons) 
            with part1 have "?part1 \<and> ?part2" ..
          }
          with res show ?thesis by auto
        qed
      qed
    qed
  } note main = this
  from main[of "[]"] show ?thesis by (cases "linear_term_impl [] t", auto)
qed

definition check_no_var :: "('f::show, 'v::show) term \<Rightarrow> shows check" where
  "check_no_var t \<equiv> check (is_Fun t) (''variable found'' +#+ shows_nl)"

lemma check_no_var_sound[simp]:
  "isOK (check_no_var t) \<longleftrightarrow> is_Fun t"
  using assms unfolding check_no_var_def by simp


definition
  check_supt :: "('f::show, 'v::show) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_supt s t \<equiv> check (s \<rhd> t) (shows t +@+ '' is not a proper subterm of '' +#+ shows s)"

definition
  check_supteq :: "('f::show, 'v::show) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_supteq s t \<equiv> check (s \<unrhd> t) (shows t +@+ '' is not a subterm of '' +#+ shows s)"

lemma isOK_check_supt[simp]:
  "isOK (check_supt s t) = ((s, t) \<in> op \<rhd>)"
unfolding check_supt_def using supt_def by auto

lemma isOK_check_supteq[simp]:
  "isOK (check_supteq s t) = ((s, t) \<in> op \<unrhd>)"
unfolding check_supteq_def using supteq_def by auto

subsection {*Additional Functions on Terms*}

fun with_arity :: "('f, 'v) term \<Rightarrow> ('f \<times> nat, 'v) term" where
  "with_arity (Var x) = Var x"
| "with_arity (Fun f ts) = Fun (f, length ts) (map with_arity ts)"

fun add_vars :: "('f, 'v) term \<Rightarrow> 'v list \<Rightarrow> 'v list" where
  "add_vars (Var x) xs = x # xs"
| "add_vars (Fun _ ts) xs = foldr add_vars ts xs"

fun add_funs :: "('f, 'v) term \<Rightarrow> 'f list \<Rightarrow> 'f list" where
  "add_funs (Var _) fs = fs"
| "add_funs (Fun f ts) fs = f # foldr add_funs ts fs"

(*FIXME: define vars and funs via add_vars and add_funs*)
(*
abbreviation vars where "vars t \<equiv> add_vars t []"
abbreviation vars where "funs t \<equiv> add_funs t []"
*)
(*FIXME: define funas via with_arity and add_funs*)

lemma with_arity_term_fold[code]:
  "with_arity = Term.fold Var (\<lambda>f ts. Fun (f, length ts) ts)"
proof
  fix t show "with_arity t = Term.fold Var (\<lambda>f ts. Fun (f, length ts) ts) t"
    by (induct t) simp_all
qed


fun flatten_term_enum :: "('f list,'v)term \<Rightarrow> ('f,'v)term list"
where "flatten_term_enum (Var x) = [Var x]"
   |  "flatten_term_enum (Fun fs ts) = (
          let lts = map flatten_term_enum ts;
              ss = concat_lists lts
            in concat (map (\<lambda> f. map (Fun f) ss) fs))"

lemma flatten_term_enum: 
  shows "set (flatten_term_enum t) = {u. instance_term u (map_term set t)}"
proof (induct t)
  case (Var x)
  show ?case (is "_ = ?R")
  proof -
    {
      fix t 
      assume "t \<in> ?R"
      hence "t = Var x" by (cases t, auto)
    } 
    thus ?thesis by auto
  qed
next
  case (Fun fs ts)
  show ?case (is "?L = ?R")
  proof -
    {
      fix i
      assume "i < length ts"
      hence "ts ! i \<in> set ts" by auto
      note Fun[OF this]
    } note ind = this
    have idL: "?L = {Fun g ss | g ss. g \<in> set fs  \<and> length ss = length ts \<and> (\<forall>i<length ts. ss ! i \<in> set (flatten_term_enum (ts ! i)))}" (is "_ = ?M1") by auto  
    let ?R1 = "{Fun f ss | f ss. f \<in> set fs \<and> length ss = length ts \<and> (\<forall> i<length ss. instance_term (ss ! i) (map_term set (ts ! i)))}"
    {
      fix u
      assume "u \<in> ?R"
      hence "u \<in> ?R1" by (cases u, auto)
    }
    hence idR: "?R = ?R1" by auto
    show ?case unfolding idL idR using ind by auto
  qed
qed

subsection {* Creating substitutions from lists *}

type_synonym ('f, 'v) substL = "('v \<times> ('f, 'v) term) list"

definition mk_subst :: "('f, 'v) substL \<Rightarrow> ('f, 'v) subst" where
  "mk_subst xts \<equiv> (\<lambda>x. case lookup x xts of
    None \<Rightarrow> Var x
  | Some t \<Rightarrow> t)"

lemma mk_subst_Nil[simp]: "mk_subst [] = Var" by (simp add: mk_subst_def)

(* TODO: perhaps implement more efficiently *)
definition mk_subst_domain :: "('f,'v)substL \<Rightarrow> ('v \<times> ('f,'v)term) list"
  where "mk_subst_domain \<sigma> \<equiv> let \<tau> = mk_subst \<sigma> in 
             remdups (filter (\<lambda>(x,t). Var x \<noteq> t) (map (\<lambda> (x,t). (x,\<tau> x)) \<sigma>))"

lemma mk_subst_domain: "set (mk_subst_domain \<sigma>) = (\<lambda> x. (x,mk_subst \<sigma> x)) ` subst_domain (mk_subst \<sigma>)" (is "?I = ?R")
proof -
  have "?I \<subseteq> ?R" unfolding mk_subst_domain_def Let_def subst_domain_def by auto
  moreover 
  {
    fix xt
    assume mem: "xt \<in> ?R"    
    obtain x t where xt: "xt = (x,t)" by force
    from mem[unfolded xt] have x: "x \<in> subst_domain (mk_subst \<sigma>)" and t: "t = mk_subst \<sigma> x" by auto
    hence "mk_subst \<sigma> x \<noteq> Var x" unfolding subst_domain_def by simp
    with t have l: "lookup x \<sigma> = Some t" and tx: "t \<noteq> Var x" 
      unfolding mk_subst_def by (cases "lookup x \<sigma>", auto)
    from lookup_sound[OF l] l t tx have "(x,t) \<in> ?I" unfolding mk_subst_domain_def Let_def
      by force
    hence "xt \<in> ?I" unfolding xt .
  }
  ultimately show ?thesis by blast
qed

lemma finite_mk_subst: "finite (subst_domain (mk_subst \<sigma>))"
proof -
  have "subst_domain (mk_subst \<sigma>) = fst ` set (mk_subst_domain \<sigma>)"
    unfolding mk_subst_domain Let_def by force
  moreover have "finite ..."
    using finite_set by auto
  ultimately show ?thesis by simp
qed

subsection {*Code Generator Setup*}

text{*The target languagte is assumed to be Haskell. Hence the following
code equations try to increase lazyness as far as possible.*}
lemma [code]:
  "add_vars = Term.fold (op #) (const (flip (foldr (op \<circ>)) id))"
    (is "_ = ?fun")
proof (intro ext)
  fix t xs show "add_vars t xs = ?fun t xs"
  proof (induct t arbitrary: xs)
    case (Var x) show ?case by simp
  next
    case (Fun f ts xs)
    from foldr_cong[of xs xs ts ts add_vars, OF _ _ this]
      have IH: "foldr add_vars ts xs = foldr ?fun ts xs" by blast
    show ?case
      by (simp add: IH foldr_map const_def del: o_apply id_apply)
  qed
qed

lemma [code]:
  "add_funs = Term.fold (const id) (\<lambda>f xs. op # f \<circ> foldr (op \<circ>) xs id)"
    (is "_ = ?fun")
proof (intro ext)
  fix t fs show "add_funs t fs = ?fun t fs"
  proof (induct t arbitrary: fs)
    case (Var x) show ?case by (simp add: const_def)
  next
    case (Fun f ts fs)
    from foldr_cong[of fs fs ts ts add_funs, OF _ _ this]
      have IH: "foldr add_funs ts fs = foldr ?fun ts fs" by blast
    show ?case
      by (simp add: IH const_def foldr_map del: o_apply id_apply)
         simp
  qed
qed

end
