(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Term_Order
imports
  Trs
  List_Order
  Monadic_Check
begin

locale compat =
  fixes  S :: "'a ars" 
  and   NS :: "'a ars" 
  assumes compat: "NS O S \<subseteq> S"
begin

lemma trCompat: "NS^* O S \<subseteq> S" using compat by (rule compat_tr_compat)

lemma trans_split_NS_S_union:
  assumes "r^* \<subseteq> (NS \<union> S)^*"
  shows "r^* \<subseteq> S O S^* O NS^* \<union> NS^*"
proof(rule subsetI2)
  fix x y
  assume "(x, y)  \<in> r^*"
  with assms have  "(x,y) \<in> (NS \<union> S)^*" by auto
  thus "(x,y) \<in> S O S^* O NS^* \<union> NS^*" using compatible_rtrancl_split[where S = S and NS = NS] compat by auto
qed
end


locale redpair = SN_rel S for
  S :: "('f,'v)trs" + fixes
  NS :: "('f,'v)trs" 
  assumes 
    wm_NS: "one_step_ctxt_closed NS"
    and subst_S: "subst_closed S"
    and subst_NS: "subst_closed NS"

locale redpair_order = redpair S NS + pre_order_pair S NS for S NS :: "('f,'v)trs" 

context redpair
begin
lemma redpair_order: "redpair_order (S^+) (NS^*)"
proof -
  let ?NS = "NS^*"
  let ?S = "S^+"
  show "redpair_order ?S ?NS"
  proof
    show "subst_closed ?NS" by (rule subst_closed_rtrancl[OF subst_NS])
  next
    show "subst_closed ?S" by (rule subst_closed_trancl[OF subst_S])
  next 
    show "one_step_ctxt_closed ?NS" by (rule rtrancl_one_step_ctxt_closed[OF wm_NS])
  next
    show "SN ?S" by (rule SN_imp_SN_trancl[OF SN])
  next
    show "refl ?NS" by (rule refl_rtrancl)
  next
    show "trans ?NS" by (rule trans_rtrancl)
  next
    show "trans ?S" by (rule trans_trancl)
  qed 
qed

lemma subst_NSS: "subst_closed (NS \<union> S)"
  using subst_S subst_NS unfolding subst_closed_def using cs_union_distr[where ?R=NS and ?S=S] by auto

lemma mono_NSS: assumes mono: "one_step_ctxt_closed S"
  shows "one_step_ctxt_closed (NS \<union> S)"
  using  mono wm_NS unfolding one_step_ctxt_closed_def by auto


lemma mono_rstep_NSS:
  assumes mono: "one_step_ctxt_closed S"
  shows "rstep (NS \<union> S) \<subseteq> NS \<union> S"
proof -
  from trs_inclusion[OF mono_NSS[OF mono] subst_NSS]
  show ?thesis by auto
qed

lemma mono_rstep_S:
  assumes mono: "one_step_ctxt_closed S"
  shows "rstep S \<subseteq> S"
proof -
  from trs_inclusion[OF mono subst_S]
  show ?thesis by auto
qed

lemma mono_rstep_NSS_rtrancl:
  assumes mono: "one_step_ctxt_closed S"
  shows "(rstep (NS \<union> S))^* \<subseteq> (NS \<union> S)^*"
  by (rule rtrancl_mono[OF mono_rstep_NSS[OF mono]])

lemma rstep_NS: "rstep NS \<subseteq> NS"
  using trs_inclusion[OF wm_NS subst_NS] by simp

lemma rstep_NS_rtrancl: "(rstep NS)^* \<subseteq> NS^*" 
  by (rule rtrancl_mono[OF rstep_NS])

lemma mNS: "ctxt_closed NS" 
  using wm_NS by (rule one_step_ctxt_closed_imp_ctxt_closed)
end

locale pre_redtriple = redpair S NS 
  for S NS :: "('f,'v)trs" +
  fixes NST :: "('f,'v)trs" 
  assumes compat_NST: "NST O S \<subseteq> S"
    and subst_NST: "subst_closed NST"

locale pre_redtriple_order = pre_redtriple S NS NST + redpair_order S NS 
  for S NS NST :: "('f,'v)trs" +
  assumes  trans_NST: "trans NST"

lemma (in pre_redtriple) pre_redtriple_order: "pre_redtriple_order (S^+) (NS^*) (NST^*)"
proof -
  let ?NS = "NS^*"
  let ?NST = "NST^*"
  let ?S = "S^+"
  from redpair_order interpret redpair_order ?S ?NS .
  show "pre_redtriple_order ?S ?NS ?NST"
  proof
    show "subst_closed ?NST" by (rule subst_closed_rtrancl[OF subst_NST])
  next
    show "?NST O ?S \<subseteq> ?S" unfolding trancl_unfold_left 
    proof (clarsimp)
      fix x y z u
      assume "(x,y) \<in> ?NST" and "(y,z) \<in> S" and zu: "(z,u) \<in> S^*"
      with compat.trCompat[unfolded compat_def, OF compat_NST] have "(x,z) \<in> S" by blast
      with zu show "(x,u) \<in> S O S^*" by auto
    qed 
  next
    show "trans ?NST" by (rule trans_rtrancl)
  qed 
qed

locale root_redtriple = pre_redtriple S NS NST for S NS NST :: "('f,'v)trs" +
  assumes 
    compat_NS_root: "(s,t) \<in> NS \<Longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> NST"
begin
lemma nrrstep_imp_NST: assumes R: "R \<subseteq> NS"
  and step: "(s,t) \<in> nrrstep R" 
  shows "(s,t) \<in> NST"
proof -
  from step[unfolded nrrstep_def']
  obtain l r C \<sigma> where one: "(l,r) \<in> R" "C \<noteq> \<box>" "s = C\<langle>l\<cdot>\<sigma>\<rangle>" "t = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
  then obtain f bef D aft where "C = More f bef D aft" by (cases C, auto)
  with one have id: "s = Fun f (bef @ D\<langle>l\<cdot>\<sigma>\<rangle> # aft)" "t = Fun f (bef @ D\<langle>r\<cdot>\<sigma>\<rangle> # aft)" by auto
  from one R have "(l,r) \<in> NS" by auto
  with subst_NS have "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> NS" unfolding subst_closed_def by auto
  with mNS have "(D\<langle>l\<cdot>\<sigma>\<rangle>,D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> NS" unfolding ctxt_closed_def by auto
  from compat_NS_root[OF this]
  show "(s,t) \<in> NST" unfolding id .
qed

lemma compat_NSs_root: 
  "(s,t) \<in> NS^* \<Longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> NST^*"
  by (induct rule: rtrancl_induct, insert compat_NS_root[of _ _ f bef aft], force+)


lemma nrrsteps_imp_NST: assumes R: "R \<subseteq> NS"
  and steps: "(s,t) \<in> (nrrstep R)^*" 
  shows "(s,t) \<in> NST^*"
  using steps
  by (induct rule: converse_rtrancl_induct, insert nrrstep_imp_NST[OF R], force+)

lemma nrrstep_compat: "nrrstep NS O S \<subseteq> S"
proof
  fix s t
  assume "(s,t) \<in> nrrstep NS O S"
  then obtain u where NS: "(s,u) \<in> nrrstep NS" and u: "(u,t) \<in> S" by auto
  from compat_NST nrrstep_imp_NST[OF subset_refl NS] u show "(s,t) \<in> S" by auto
qed

lemma compat_NSTs: "(s,t) \<in> NST^* \<Longrightarrow> (t,u) \<in> S \<Longrightarrow> (s,u) \<in> S"
  by (induct rule: converse_rtrancl_induct, insert compat_NST, auto)
end

sublocale root_redtriple \<subseteq> nrr: compat S "nrrstep NS \<union> NST" 
  by (unfold_locales, insert nrrstep_compat compat_NST, auto)

locale redtriple = pre_redtriple S NS NST + compat_pair S NS for S NS NST :: "('f,'v)trs"     

sublocale redtriple \<subseteq> both: compat S "NS \<union> NST"
  by (unfold_locales, insert compat_NS_S compat_NST, auto)

sublocale redtriple \<subseteq> NS: compat S NS
  by (unfold_locales, insert compat_NS_S compat_NST, auto)


locale redtriple_order = redtriple S NS NST + pre_redtriple_order S NS NST + order_pair S NS for S NS NST :: "('f,'v)trs" +
  assumes refl_NST: "refl NST"

sublocale redtriple_order \<subseteq> SN_order_pair  
  by (unfold_locales, rule SN)

locale root_redtriple_order = root_redtriple S NS NST + pre_redtriple_order S NS NST for S NS NST :: "('f,'v)trs"

lemma (in redtriple) redtriple_order: "redtriple_order (S^+) (NS^*) (NST^*)"
proof -
  let ?NS = "NS^*"
  let ?NST = "NST^*"
  let ?S = "S^+"
  from pre_redtriple_order interpret pre_redtriple_order ?S ?NS ?NST .
  show ?thesis
  proof
    show "?NS O ?S \<subseteq> ?S" unfolding trancl_unfold_left 
    proof (clarsimp)
      fix x y z u
      assume "(x,y) \<in> ?NS" and "(y,z) \<in> S" and zu: "(z,u) \<in> S^*"
      with trCompat have "(x,z) \<in> S" by auto
      with zu show "(x,u) \<in> S O S^*" by auto
    qed
  next
    show "?S O ?NS \<subseteq> ?S" unfolding trancl_unfold_right 
    proof (clarsimp)
      fix x y z u
      assume xy: "(x,y) \<in> S^*" and yz: "(y,z) \<in> S" and zu: "(z,u) \<in> NS^*"
      from zu have "(y,u) \<in> S"
        by (induct rule: rtrancl_induct, insert yz compat_S_NS, auto)
      with xy show "(x,u) \<in> S^* O S" by auto
    qed
  next
    show "refl ?NST" by (rule refl_rtrancl)
  qed
qed

lemma (in root_redtriple) root_redtriple_order: "root_redtriple_order (S^+) (NS^*) (NST^*)"
proof -
  let ?NS = "NS^*"
  let ?NST = "NST^*"
  let ?S = "S^+"
  from pre_redtriple_order interpret pre_redtriple_order ?S ?NS ?NST .
  show ?thesis
  proof
    fix s t f bef aft u
    assume st: "(s,t) \<in> ?NS" 
    have "(s,t) \<in> (rstep NS)^*" by (rule set_mp[OF rtrancl_mono st], auto)
    from nrrsteps_imp_NST[OF _ rsteps_ctxt_imp_nrrsteps[OF this, of "More f bef \<box> aft"]]
    show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> ?NST" by auto
  qed
qed

sublocale redtriple \<subseteq> root_redtriple
proof (unfold_locales)
  fix s t f bef aft u
  assume NS: "(s,t) \<in> NS" 
  from one_step_ctxt_closed_one[OF wm_NS NS, of f bef aft]
  have "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> NS" .
  thus "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> NST" 
oops


lemma (in redpair_order) all_ctxt_closed: "all_ctxt_closed NS"
  by (rule tr_one_imp_all_ctxt_closed[OF trans_NS refl_NS wm_NS])


fun ce_trs :: "('f \<times> 'f \<times> nat) \<Rightarrow> ('f,'v)trs"
where "ce_trs (c,d,n) = 
  {(Fun c (t # s # replicate n (Fun d [])), t) | t s. True}
  \<union> {(Fun c (t # s # replicate n (Fun d [])), s) | t s. True}"

fun comb :: "('f \<times> 'f \<times> nat) \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term"
where "comb (c,d,_) [] = Fun d []"
   |  "comb (c,d,n) (t # ts) = Fun c (t # comb (c,d,n) ts # replicate n (Fun d []))"

lemma ce_trs_sound: "t \<in> set ts \<Longrightarrow> (comb cn ts, t) \<in> (rstep (ce_trs cn))^+"
using assms 
proof (induct ts, simp)
  case (Cons s ss)
  show ?case 
  proof (cases cn)
    case (fields c d n)
    from Cons show ?thesis 
    proof (cases "s = t")
      case True 
      with fields have "(comb cn (s # ss),s) \<in> ce_trs cn" by simp
      hence "(comb cn (s # ss),t) \<in> rstep (ce_trs cn)" using True by blast
      thus ?thesis using True by force
    next
      case False
      hence ind: "(comb cn ss, t) \<in> (rstep (ce_trs cn))^+" using Cons by auto
      from fields have "(comb cn (s # ss),comb cn ss) \<in> ce_trs cn" by simp
      hence "(comb cn (s # ss),comb cn ss) \<in> rstep (ce_trs cn)" by blast 
      with ind show ?thesis by auto
    qed
  qed
qed

declare ce_trs.simps[simp del] comb.simps[simp del]


type_synonym 'f af = "('f \<times> nat) \<Rightarrow> nat \<Rightarrow> bool"

fun af_regarded_pos :: "'f af \<Rightarrow> ('f,'v)term \<Rightarrow> pos \<Rightarrow> bool"
  where "af_regarded_pos \<pi> t \<epsilon> = True"
      | "af_regarded_pos \<pi> (Fun f ts) (PCons i p) = (i < length ts \<and> \<pi> (f,length ts) i \<and> af_regarded_pos \<pi> (ts ! i) p)"
      | "af_regarded_pos \<pi> (Var _) (PCons i p) = False"

definition af_compatible :: "'f af \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
where "af_compatible \<pi> ord \<equiv>
   (\<forall> f bef s t aft. \<pi> (f, (Suc (length bef + length aft))) (length bef) \<or>
   (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> ord)"

lemma af_compatible_af_regarded_ctxt: assumes af: "af_compatible \<pi> ord"
  and one: "one_step_ctxt_closed ord"
  and not: "\<not> af_regarded_pos \<pi> (C\<langle>u\<rangle>) (hole_pos C)"
  shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> ord"
  using not
proof (induct C)
  case Hole
  thus ?case by simp
next
  case (More f bef C aft)
  let ?i = "length bef"
  let ?n = "Suc (?i + length aft)"
  let ?pi = "\<pi> (f,?n) ?i"
  show ?case 
  proof (cases ?pi)
    case True
    with More(2) have "\<not> af_regarded_pos \<pi> (C\<langle>u\<rangle>) (hole_pos C)" by simp
    from More(1)[OF this] have "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> ord" .
    from one_step_ctxt_closed_one[OF one this]
    show ?thesis by auto
  next
    case False
    with af[unfolded af_compatible_def, rule_format, of f bef aft]
    show ?thesis by auto
  qed
qed

definition full_af :: "'f af" where "full_af fn i \<equiv> True"

lemma full_af: "af_compatible full_af r" unfolding af_compatible_def full_af_def by auto

lemma af_regarded_poss: "af_regarded_pos \<pi> t p \<Longrightarrow> p \<in> poss t"
proof (induct p arbitrary: t)
  case Empty
  show ?case by simp
next
  case (PCons i p t)
  show ?case
  proof (cases t)
    case (Var x)
    show ?thesis using PCons(2) unfolding Var by simp
  next
    case (Fun f ts)
    show ?thesis using PCons(2) PCons(1)[of "ts ! i"] unfolding Fun by auto
  qed
qed

lemma af_regarded_full: "af_regarded_pos full_af t p = (p \<in> poss t)"
proof (induct p arbitrary: t)
  case Empty
  show ?case by simp
next
  case (PCons i p t)
  show ?case
  proof (cases t)
    case (Var x)
    show ?thesis unfolding Var by simp
  next
    case (Fun f ts)
    show ?thesis using PCons[of "ts ! i"] unfolding Fun by (auto simp: full_af_def)
  qed
qed

lemma af_steps_imp_orient: assumes tran: "trans r" 
  and refl: "refl r"
  and one: "one_step_ctxt_closed r" 
  and len: "length ts = length (ss :: ('f,'v)term list)"
  and steps: "\<forall>i<length ts. p i \<longrightarrow> (ts ! i, ss ! i) \<in> r"
  and compat: "\<forall> bef s t aft. ((length ts = Suc (length bef + length aft)) \<longrightarrow> (p (length bef) \<or> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r))"
  shows "(Fun f ts, Fun f ss) \<in> r" 
proof -
  from trans_refl_imp_rtrancl_id[OF tran refl] have r: "r^* = r" by auto
  let ?rel = "\<lambda> i. if p i then r else UNIV"
  have "(Fun f ts, Fun f ss) \<in> r^*"
  proof (rule args_steps_imp_steps_gen[OF _ len, of ?rel])
    fix i
    assume i: "i < length ss"
    with len steps show "(ts ! i, ss ! i) \<in> (?rel i)^*" by auto
  next
    fix bef aft :: "('f,'v)term list" and s t 
    assume rel: "(s,t) \<in> ?rel (length bef)" and len': "length ss = Suc (length bef + length aft)"
    from compat[rule_format, OF len'[unfolded len[symmetric]]]
    have "p (length bef) \<or> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r" .
    thus "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r^*"
    proof
      assume "p (length bef)"
      with rel have "(s,t) \<in> r" by auto
      with one[unfolded one_step_ctxt_closed_def] show ?thesis by auto
    qed auto
  qed
  thus ?thesis unfolding r .
qed

locale af_redpair = redpair  S NS for S NS :: "('f,'v)trs"+
  fixes \<pi> :: "'f af"
  assumes af_compat: "af_compatible \<pi> NS"
begin
lemma af_redpair_order: "af_redpair (S^+) (NS^*) \<pi>"
proof -
  let ?NS = "NS^*"
  let ?S = "S^+"
  from redpair_order interpret redpair_order ?S ?NS .
  show ?thesis
  proof
    show "af_compatible \<pi> ?NS" using af_compat
      unfolding af_compatible_def by blast
  qed
qed
end

locale ce_redpair = redpair S NS for S NS :: "('f,'v)trs"+ 
  assumes NS_ce_compat: "\<exists> n. \<forall> m. m \<ge> n \<longrightarrow> (\<forall> c d. ce_trs (c,d,m) \<subseteq> NS)"
begin 
lemma ce_redpair_order: "ce_redpair (S^+) (NS^*)"
proof -
  let ?S = "S^+"
  let ?NS = "NS^*"
  from redpair_order interpret redpair_order ?S ?NS .
  show ?thesis using NS_ce_compat
    by (unfold_locales, blast)
qed
end

locale ce_af_redpair = ce_redpair S NS + af_redpair S NS \<pi>
  for S NS :: "('f,'v)trs" and \<pi>

locale ce_redtriple = redtriple  S NS NST + ce_redpair S NS for S NS NST :: "('f,'v)trs"
begin
lemma ce_redtriple_order: "ce_redtriple (S^+) (NS^*) (NST^*)"
proof -
  let ?S = "S^+"
  let ?NS = "NS^*"
  let ?NST = "NST^*"
  from redtriple_order ce_redpair_order interpret redtriple_order ?S ?NS ?NST + ce_redpair ?S ?NS .
  show ?thesis by (unfold_locales)
qed
end


locale ce_af_root_redtriple = root_redtriple S NS NST + ce_af_redpair S NS \<pi> for S NS NST :: "('f,'v)trs" and \<pi> +
  fixes \<pi>' :: "'f af"
  assumes af_compat': "af_compatible \<pi>' NST"

locale ce_af_root_redtriple_order = root_redtriple_order S NS NST + ce_af_root_redtriple S NS NST \<pi> \<pi>' for S NS NST :: "('f,'v)trs" and \<pi> \<pi>'

locale ce_af_redtriple = ce_redtriple S NS NST + ce_af_redpair S NS \<pi> for S NS NST :: "('f,'v)trs" and \<pi>

locale ce_af_redtriple_order = ce_redtriple S NS NST + ce_af_redpair S NS \<pi> + redtriple_order S NS NST for S NS NST :: "('f,'v)trs" and \<pi>

sublocale ce_af_redtriple_order \<subseteq> ce_af_redtriple ..

lemma (in ce_af_redtriple) ce_af_redtriple_order: "ce_af_redtriple_order (S^+) (NS^*) (NST^*) \<pi>"
proof -
  let ?S = "S^+"
  let ?NS = "NS^*"
  let ?NST = "NST^*"
  from redtriple_order ce_redpair_order af_redpair_order interpret redtriple_order ?S ?NS ?NST + ce_redpair ?S ?NS + af_redpair ?S ?NS \<pi> .
  show ?thesis by (unfold_locales)
qed


locale mono_ce_redtriple = redtriple S NS NST + ce_redpair S NS for S NS NST :: "('f,'v)trs" + 
  assumes S_ce_compat: "\<exists> n. \<forall> m. m \<ge> n \<longrightarrow> (\<forall> c d. ce_trs (c,d,m) \<subseteq> S)"
begin 
lemma mono_ce_redtriple_order: "mono_ce_redtriple (S^+) (NS^*) (NST^*)"
proof -
  let ?S = "S^+"
  let ?NS = "NS^*"
  let ?NST = "NST^*"
  have subset: "\<And> x. x \<in> S \<Longrightarrow> x \<in> ?S" by auto
  from redtriple_order ce_redpair_order interpret redtriple_order ?S ?NS ?NST + ce_redpair ?S ?NS .
  show ?thesis using S_ce_compat subset
    by (unfold_locales, blast)
qed
end
  

locale mono_ce_af_redtriple = ce_af_redtriple S NS NST \<pi> + mono_ce_redtriple S NS NST for S NS NST :: "('f,'v)trs" and \<pi>

locale mono_ce_af_redtriple_order = mono_ce_af_redtriple S NS NST \<pi> + redtriple_order S NS NST for S NS NST :: "('f,'v)trs" and \<pi>

sublocale mono_ce_af_redtriple_order \<subseteq> ce_af_redtriple_order ..

lemma (in mono_ce_af_redtriple)  mono_ce_af_redtriple_order: "mono_ce_af_redtriple_order (S^+) (NS^*) (NST^*) \<pi>"
proof -
  let ?S = "S^+"
  let ?NS = "NS^*"
  let ?NST = "NST^*"
  from ce_af_redtriple_order mono_ce_redtriple_order interpret ce_af_redtriple_order ?S ?NS ?NST \<pi> + mono_ce_redtriple ?S ?NS ?NST .
  show ?thesis ..
qed


end
