(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Tree_Automata_Checker
imports Xml Tree_Automata_Impl Linear_Orders
begin

(* TODO: several parts are just copied from CPF_Parser and
  these should be shared *)

fun
  var :: "('a, string) term xmlt"
where
  "var x = (Xml.text ''var'' x \<guillemotright>= (return \<circ> Var))"
hide_const (open) var

function
  "term" :: "'a xmlt \<Rightarrow> ('a, string) term xmlt"
where
  "term xml2name x = 
     xml_options [(''var'', Tree_Automata_Checker.var), (''funapp'', xml1many_elements ''funapp'' xml2name (xml1element ''arg'' (term xml2name) id) Fun)] x"  
by (pat_completeness) auto
hide_const (open) "term"

definition
  rule :: "'a xmlt \<Rightarrow> ('a, string) rule xmlt"
where
  "rule xml2name = xml2elements ''rule'' (xml1element ''lhs'' (Tree_Automata_Checker.term xml2name) id) (xml1element ''rhs'' (Tree_Automata_Checker.term xml2name) id) Pair"
hide_const (open) rule 

definition
  rules :: "'a xmlt \<Rightarrow> ('a,string) rules xmlt"
where
  "rules xml2name = xml_many_elements ''rules'' (Tree_Automata_Checker.rule xml2name) id"
hide_const (open) rules

definition state :: "string xmlt" where
  "state = Xml.text ''state''"


definition final_states :: "string list xmlt" where
  "final_states = xml_many_elements ''finalStates'' state id"

(* TODO: move this change to CPF_Parser (optional height) *)
definition transition_lhs :: "'a xmlt \<Rightarrow> ('a \<times> string list + string) xmlt" where
  "transition_lhs xml2name x = (if (case x of (XML _ _ cs _) \<Rightarrow> length cs = 1 \<and> (
                                    case hd cs of (XML name _ _ _) \<Rightarrow> name = ''state''
            )) then 
    (xml1element ''lhs'' state Inr) x
   else
    xml1many_elements ''lhs''
    (xml2name)
    (state)
    (\<lambda>f qs. Inl (f, qs)) x)"

definition transition :: "'a xmlt \<Rightarrow> ((string, 'a) ta_rule + (string \<times> string)) xmlt" where
  "transition xml2name = xml2elements ''transition''
    (transition_lhs xml2name)
    (xml1element ''rhs'' state id)
    (\<lambda> lhs q. case lhs of Inl (f, qs) \<Rightarrow> Inl (TA_rule f qs q) | Inr q' \<Rightarrow> Inr (q',q))"

definition transitions :: "'a xmlt \<Rightarrow> (((string, 'a) ta_rule list) \<times> ((string \<times> string)list)) xmlt" where
  "transitions xml2name = xml_many_elements ''transitions'' (transition xml2name)
     (\<lambda> rls. let (rules, eps) = partition (\<lambda> rl. case rl of Inl _ \<Rightarrow> True | Inr _ \<Rightarrow> False) rls;
                  ruls = map (\<lambda> rl. case rl of Inl r \<Rightarrow> r) rules;
                  ep   = map (\<lambda> rl. case rl of Inr e \<Rightarrow> e) eps
                in (ruls,ep))"


definition tree_automaton :: "'a xmlt \<Rightarrow> (string, 'a) tree_automaton xmlt" where
  "tree_automaton xml2name \<equiv> xml2elements ''treeAutomaton''
    final_states
    (transitions xml2name)
    (\<lambda> f (t, eps). Tree_Automaton f t eps)"

definition
  xml_tree_aut_problem ::
    "'a xmlt 
     \<Rightarrow> (string, 'a) tree_aut_problem xmlt"
where "xml_tree_aut_problem xml2name = xml3elements ''treeAutomataProblem'' 
       (tree_automaton xml2name)
       (xml1element ''trs'' (Tree_Automata_Checker.rules xml2name) id)
       (tree_automaton xml2name)
       Tree_Automata_Impl.tree_aut_problem"

primrec
  xmldoc2cert_problem :: "xmldoc \<Rightarrow> string + (string,string)tree_aut_problem"
where
  "xmldoc2cert_problem (XMLDOC header xml) = 
     xml_tree_aut_problem (Xml.text ''name'') xml"

fun
  parse_cert_problem :: "string \<Rightarrow> string + (string,string)tree_aut_problem"
where
  "parse_cert_problem s = (doc_of_string s \<guillemotright>= xmldoc2cert_problem)"

datatype cert_result = Certified | Unsupported string | Reject string 

fun certify_proof :: "string \<Rightarrow> cert_result"
where "certify_proof problemString  = (
  case parse_cert_problem problemString of
     Inl err \<Rightarrow> Unsupported (''error while parsing'' @ [newline] @ err)
   | Inr tap \<Rightarrow> case (certify_tree_aut_problem tap) of
                         Inr () \<Rightarrow> Certified 
                      |  Inl err  \<Rightarrow> Reject (err []))"

end



