theory Usable_Rules_2
imports QDP_Framework Trs_Impl
begin

fun ur_term :: "('f,'v)rules \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)rules"
  where "ur_term R (Var x) = []"
     |  "ur_term R (Fun f ts) = remdups (concat (map (ur_term R) ts)
  @ filter (\<lambda> (l,r). root l = Some (f,length ts)) R)"

lemma ur_term_subset_R: "set (ur_term R t) \<subseteq> set R"
  by (induct t, auto)


function ur_dp' :: "('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules"
where "ur_dp' R U = (let U' = remdups (concat (map (\<lambda> (s,t). (s,t) # ur_term R t) U)) in
       if set U' \<subseteq> set U then U else ur_dp' R U')"
  by pat_completeness auto

termination
proof -
  let ?M = "\<lambda> (R,U). size (filter (\<lambda> x. x \<notin> set U) R)"
  show ?thesis
  proof
    show "wf (measure ?M)" by simp
  next
    fix R U x
    assume x: "x = remdups (concat (map (\<lambda>(s,t). (s,t) # ur_term R t) U))" (is "_ = ?U")
      and neq: "\<not> set x \<subseteq> set U"
    let ?lsmall = "filter (\<lambda> x. x \<notin> set ?U) R"
    let ?llarge = "filter (\<lambda> x. x \<notin> set U) R"
    from neq x have neq: "set U \<subset> set ?U" by auto
    then obtain lr where lr: "lr \<in> set ?U - set U" by fast
    then obtain t where urt: "lr \<in> set (ur_term R t)" by auto
    hence lrR: "lr \<in> set R" using ur_term_subset_R by auto
    show "((R,x), (R,U)) \<in> measure ?M"
      unfolding x
      unfolding in_measure
      unfolding split
    proof -      
      let ?P = "\<lambda> lr. lr \<notin> set ?U - set U"
      let ?filt = "filter ?P ?llarge"
      have id: "?lsmall = ?filt" unfolding filter_filter 
        by (rule filter_cong, auto)
      show "length ?lsmall < length ?llarge"
        unfolding id
      proof (rule length_filter_less)
        show "lr \<in> set ?llarge" using lr lrR by auto
      next
        show "\<not> ?P lr" using lr by simp
      qed
    qed
  qed
qed

declare ur_dp'.simps[simp del]

lemma ur_dp'_mono: "set U \<subseteq> set (ur_dp' R U)"
proof (induct U rule: ur_dp'.induct[of "\<lambda> R U. set U \<subseteq> set (ur_dp' R U)"])
  case (1 R U)
  let ?U = "remdups (concat (map (\<lambda>(s, t). (s, t) # ur_term R t) U))"
  note unf = ur_dp'.simps[of R U, unfolded Let_def]
  show ?case 
  proof (cases "set ?U \<subseteq> set U")
    case True
    show ?thesis unfolding unf using True by simp
  next
    case False
    hence id: "set ?U \<subseteq> set U = False" by simp
    show ?thesis unfolding unf id if_False
    proof -
      have "set U \<subseteq> set ?U" by auto
      also have "... \<subseteq> set (ur_dp' R ?U)"
        by (rule 1, insert False, auto)
      finally show "set U \<subseteq> set (ur_dp' R ?U)" .
    qed
  qed
qed

lemma ur_dp'_ur_closed: assumes mem: "(l',r') \<in> set (ur_dp' R U)"
  shows "set (ur_term R r') \<subseteq> set (ur_dp' R U)"
proof (induct rule:
   ur_dp'.induct[of "\<lambda> R U. (l',r') \<in> set (ur_dp' R U) \<longrightarrow> set (ur_term R r') \<subseteq> set (ur_dp' R U)", 
    rule_format, OF _ mem])
  case (1 R U)
  note IH = 1(1)
  let ?U = "remdups (concat (map (\<lambda>(s,t). (s,t) # ur_term R t) U))"
  show ?case
  proof (cases "set ?U \<subseteq> set U")
    case False
    note IH = IH[OF refl False]
    have id: "ur_dp' R ?U = ur_dp' R U" unfolding ur_dp'.simps[where U = U] Let_def
      using False by simp
    show ?thesis using IH unfolding id using 1(2) by simp
  next
    case True
    hence id1: "set ?U = set U" by auto
    have id2: "ur_dp' R U = U" unfolding ur_dp'.simps[where U = U] Let_def
      using True by simp
    have "set (ur_term R r') \<subseteq> set ?U" using 1(2) unfolding id2 by auto
    also have "... = set U" unfolding id1 ..
    also have "... = set (ur_dp' R U)" unfolding id2 ..
    finally show ?thesis .
  qed
qed
  

fun ur_dp :: "('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules"
where "ur_dp P R = ur_dp' R (remdups (concat (map (\<lambda> (s,t). ur_term R t) P)))" 

inductive_set 
  usable_rules for P R
  where fromP: "(s,t) \<in> P \<Longrightarrow> lr \<in> set (ur_term R t) \<Longrightarrow> lr \<in> usable_rules P R"
     |  fromU: "(l',r') \<in> usable_rules P R \<Longrightarrow> lr \<in> set (ur_term R r') \<Longrightarrow> lr \<in> usable_rules P R"

lemma ur_dp'_subset_usable_rules: "set U \<subseteq> usable_rules P R \<Longrightarrow> set (ur_dp' R U) \<subseteq> usable_rules P R"
proof (induct U rule: ur_dp'.induct)
  case (1 R U)
  let ?U = "remdups (concat (map (\<lambda>(s,t). (s,t) # ur_term R t) U))"
  let ?UPR = "usable_rules P R"
  show ?case
  proof (cases "set ?U \<subseteq> set U")
    case True
    hence "set (ur_dp' R U) = set U" unfolding ur_dp'.simps[where U = U] Let_def by simp
    thus ?thesis using 1(2) by simp
  next
    case False
    hence id: "set (ur_dp' R ?U) = set (ur_dp' R U)" unfolding ur_dp'.simps[where U = U] Let_def by simp
    note IH = 1(1)    
    show ?thesis
    proof (rule IH[OF refl False, unfolded id])
      show "set ?U \<subseteq> ?UPR"
      proof
        fix l r
        assume "(l,r) \<in> set ?U"
        hence "(l,r) \<in> set U \<or> (\<exists> l' r'. (l',r') \<in> set U \<and> (l,r) \<in> set (ur_term R r'))" by auto
        thus "(l,r) \<in> ?UPR"
        proof
          assume "(l,r) \<in> set U"
          with 1(2) show ?thesis by auto
        next
          assume "\<exists> l' r'. (l',r') \<in> set U \<and> (l,r) \<in> set (ur_term R r')" 
          then obtain l' r' where lr': "(l',r') \<in> set U" and lr: "(l,r) \<in> set (ur_term R r')" by auto
          show ?thesis 
            by (rule fromU[OF _ lr], insert 1(2) lr', auto)
        qed
      qed
    qed
  qed
qed
  

lemma ur_dp: "set (ur_dp P R) = usable_rules (set P) R" (is "?L = ?R")
proof -
  let ?U_init = "(remdups (concat (map (\<lambda>(s, y). ur_term R y) P)))"
  {
    fix lr
    assume lr: "lr \<in> ?R"
    hence "lr \<in> ?L"
    proof (induct rule: usable_rules.induct)
      case (fromP s t lr)
      from fromP have "lr \<in> set ?U_init" by auto
      hence "lr \<in> set (ur_dp' R ?U_init)" using ur_dp'_mono by blast
      thus ?case unfolding ur_dp.simps .
    next
      case (fromU l' r' lr)
      note IH = fromU(2)
      from IH have "(l',r') \<in> set (ur_dp' R ?U_init)" by simp
      from ur_dp'_ur_closed[OF this] fromU(3) have "lr \<in> set (ur_dp' R ?U_init)" by auto      
      thus ?case by simp
    qed
  } 
  moreover 
  {
    have "?L \<subseteq> ?R" unfolding ur_dp.simps
    proof (rule ur_dp'_subset_usable_rules)
      show "set ?U_init \<subseteq> ?R"
      proof
        fix l r
        assume lr: "(l,r) \<in> set ?U_init"
        then obtain l' r' where lr': "(l',r') \<in> set P" and lr: "(l,r) \<in> set (ur_term R r')" by auto
        show "(l,r) \<in> ?R"
          by (rule fromP[OF lr' lr])
      qed
    qed
  }
  ultimately show ?thesis by blast
qed
    

end