(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Util2
  imports Main
begin

definition replace :: "'a \<Rightarrow> 'a set \<Rightarrow> 'a set \<Rightarrow> 'a set"
  where "replace a bs M \<equiv> if a \<in> M then M - {a} \<union> bs else M"

definition replace_impl :: "'a \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
  where "replace_impl a bs M \<equiv> if a \<in> set M then bs @ filter (\<lambda> b. b \<noteq> a) M else M"

lemma replace_impl[simp]: "set (replace_impl a bs M) = replace a (set bs) (set M)"
  unfolding replace_impl_def replace_def by auto


(* TODO: this function is currently also available in RBT_Map_Set_Extension in AFP/Transive_Closure *)
(* however, the import of the Collection Framework via RBT_Map_Set_Extension has some undesired 
   behaviour elsewhere
   \<Rightarrow> for next AFP update, move sorted_list_subset out of Transitive_Closure into Util ! *)
fun sorted_list_subset :: "'a :: linorder list \<Rightarrow> 'a list \<Rightarrow> 'a option"
  where "sorted_list_subset (a # as) (b # bs) = 
            (if a = b then sorted_list_subset as (b # bs)
        else if a > b then sorted_list_subset (a # as) bs
        else Some a)"
    |  "sorted_list_subset [] _ = None"
    |  "sorted_list_subset (a # _) [] = Some a"
   
lemma sorted_list_subset: assumes "sorted as" and "sorted bs"
  shows "(sorted_list_subset as bs = None) = (set as \<subseteq> set bs)"
  using assms 
proof (induct rule: sorted_list_subset.induct)
  case (2 bs)
  thus ?case by auto
next
  case (3 a as)
  thus ?case by auto
next
  case (1 a as b bs)
  from 1(3) have sas: "sorted as" and a: "\<And> a'. a' \<in> set as \<Longrightarrow> a \<le> a'" unfolding linorder_class.sorted.simps[of "a # as"] by auto
  from 1(4) have sbs: "sorted bs" and b: "\<And> b'. b' \<in> set bs \<Longrightarrow> b \<le> b'" unfolding linorder_class.sorted.simps[of "b # bs"] by auto
  show ?case
  proof (cases "a = b")
    case True
    from 1(1)[OF this sas 1(4)] True show ?thesis by auto
  next
    case False note oFalse = this
    show ?thesis 
    proof (cases "a > b")
      case True
      with a b have "b \<notin> set as" by force
      with 1(2)[OF False True 1(3) sbs] False True show ?thesis by auto
    next
      case False
      with oFalse have "a < b" by auto
      with a b have "a \<notin> set bs" by force
      with oFalse False show ?thesis by auto
    qed
  qed
qed

text {* functions for removing the i-th element of a list and adjusting an index for this removal *}
definition remove_idx :: "nat \<Rightarrow> 'a list \<Rightarrow> 'a list"
  where "remove_idx i xs \<equiv> take i xs @ drop (Suc i) xs"

lemma remove_idx_len: assumes i: "i < length xs"
  shows "length xs = Suc (length (remove_idx i xs))"
proof -
  show ?thesis unfolding arg_cong[where f = "length", OF id_take_nth_drop[OF i]]
    unfolding remove_idx_def by simp
qed

definition adjust_idx :: "nat \<Rightarrow> nat \<Rightarrow> nat"
  where "adjust_idx i j \<equiv> (if j < i then j else (Suc j))"

definition adjust_idx_rev :: "nat \<Rightarrow> nat \<Rightarrow> nat"
  where "adjust_idx_rev i j \<equiv> (if j < i then j else j - Suc 0)"

lemma adjust_idx_rev1: "adjust_idx_rev i (adjust_idx i j) = j"
  unfolding adjust_idx_def adjust_idx_rev_def by (cases "i < j", auto)

lemma adjust_idx_rev2: assumes "j \<noteq> i" shows "adjust_idx i (adjust_idx_rev i j) = j"
  unfolding adjust_idx_def adjust_idx_rev_def using assms by (cases "i < j", auto)

lemma adjust_idx_i: "adjust_idx i j \<noteq> i" unfolding adjust_idx_def by (cases "j < i", auto)

lemma adjust_idx_nth: assumes i: "i < length xs"
  shows "remove_idx i xs ! j = xs ! adjust_idx i j" (is "?l = ?r")
proof -
  let ?j = "adjust_idx i j"
  from i have ltake: "length (take i xs) = i" by simp
  note nth_xs = arg_cong[where f = "\<lambda> xs. xs ! ?j", OF id_take_nth_drop[OF i], unfolded nth_append ltake]
  show ?thesis
  proof (cases "j < i")    
    case True
    hence j: "?j = j" unfolding adjust_idx_def by simp
    show ?thesis unfolding nth_xs unfolding j remove_idx_def nth_append ltake
      using True by simp
  next
    case False    
    hence j: "?j = Suc j" unfolding adjust_idx_def by simp
    from i have lxs: "min (length xs) i = i" by simp
    show ?thesis unfolding nth_xs unfolding j remove_idx_def nth_append
      using False by (simp add: lxs)
  qed
qed

lemma adjust_idx_rev_nth: assumes i: "i < length xs"
  and ji: "j \<noteq> i"
  shows "remove_idx i xs ! adjust_idx_rev i j = xs ! j" (is "?l = ?r")
proof -
  let ?j = "adjust_idx_rev i j"
  from i have ltake: "length (take i xs) = i" by simp
  note nth_xs = arg_cong[where f = "\<lambda> xs. xs ! j", OF id_take_nth_drop[OF i], unfolded nth_append ltake]
  show ?thesis
  proof (cases "j < i")    
    case True
    hence j: "?j = j" unfolding adjust_idx_rev_def by simp
    show ?thesis unfolding nth_xs unfolding j remove_idx_def nth_append ltake
      using True by simp
  next
    case False
    with ji have ji: "j > i" by auto
    hence j: "?j = j - Suc 0" unfolding adjust_idx_rev_def by simp
    show ?thesis unfolding nth_xs unfolding j remove_idx_def nth_append ltake
      using ji by auto
  qed
qed

lemma adjust_idx_length: assumes i: "i < length xs"
  and j: "j < length (remove_idx i xs)"
  shows "adjust_idx i j < length xs"
  using j unfolding remove_idx_len[OF i] adjust_idx_def by (cases "j < i", auto)

lemma adjust_idx_rev_length: assumes i: "i < length xs"
  and j: "j < length xs"
  and ji: "j \<noteq> i"
  shows "adjust_idx_rev i j < length (remove_idx i xs)"
  using j i ji unfolding remove_idx_len[OF i] adjust_idx_rev_def by (cases "j < i", auto)

lemma [code_unfold]: "(set xs = set ys) = (set xs \<subseteq> set ys \<and> set ys \<subseteq> set xs)"
  by auto

lemma zip_size_aux: "list_size (size o snd) (zip ts ls) \<le> (list_size size ls)"
proof (induct ls arbitrary: ts)
  case Nil thus ?case by auto
next
  case (Cons l ls ts)
  thus ?case by (cases ts, auto)
qed


end
