(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Xml
imports XML_Parser
begin

subsection {* Pretty Printing of XML nodes and documents *}

instantiation xml :: "show" begin
  definition
    shows_attr :: "string \<times> string \<Rightarrow> shows"
  where
    "shows_attr av = (shows (fst av) +@+ ''='' +#+ shows_string (dquote # (snd av) @ [dquote]))"

  definition
    shows_attrs :: "(string\<times>string)list \<Rightarrow> shows"
  where
    "shows_attrs as = shows_map (\<lambda>a. '' '' +#+ shows_attr a) as"

  fun
    shows_XML_indent :: "string \<Rightarrow> nat \<Rightarrow> xml \<Rightarrow> shows"
  where
    "shows_XML_indent ind i (XML n a c t) = (
    ind +#+ ''<'' +#+ shows n +@+ shows_attrs a +@+ (
      if c = [] \<and> t = None then shows_string ''/>'' else (
        ''>'' +#+ (
          if c \<noteq> [] then (
            shows_map (shows_XML_indent (ind @ replicate i (CHR '' '')) i) c +@+ shows_string ind
          ) else id
        ) +@+ (if t = None then id else shows_string(the t)) +@+
          ''</'' +#+ shows n +@+ shows_string ''>''
      )
    )
  )"
  definition "shows_prec (d::nat) xml \<equiv> shows_XML_indent [newline] 2 xml"
  definition "shows_list (xs::xml list) \<equiv> shows_list_aux shows xs"
  
  lemma shows_attr_assoc: "(s +#+ shows_attr av) r @ t = (s +#+ shows_attr av) (r@t)"
    unfolding shows_attr_def by (cases av) (auto simp: show_defs o_def dquote_def)
  lemma shows_attr_assoc_elt: "\<forall>av\<in>set avs. (s +#+ shows_attr av) r @ t = (s +#+ shows_attr av) (r@t)"
    using shows_attr_assoc by best
  lemma shows_attrs_assoc: "shows_attrs as r @ s = shows_attrs as (r @ s)"
    unfolding shows_attrs_def shows_map_assoc[OF shows_attr_assoc_elt[of as "'' ''"]] by simp
  lemma assoc_xml': "shows_XML_indent ind i xml r @ s = shows_XML_indent ind i xml (r @ s)"
  proof (induct xml arbitrary: ind i r s)
    let ?elt = "shows_XML_indent"
    case (XML n a cs t ind i)
    hence elt: "\<And>r s.\<forall>c\<in>set cs. ?elt (ind @ replicate i CHR '' '') i c r @ s = ?elt (ind @ replicate i CHR '' '') i c (r@s)" by auto
    have IH: "\<And>r s. shows_map (?elt (ind @ replicate i CHR '' '') i) cs r @ s = shows_map (?elt (ind @ replicate i CHR '' '') i) cs (r@s)"
      by (rule shows_map_assoc[OF elt])
    show ?case
    proof (cases "cs = []")
      case True show ?thesis
      proof (cases t)
        case None
        with True have empty: "cs = [] \<and> t = None" by simp
        show ?thesis by (auto simp: shows_attrs_assoc if_P[OF empty])
      next
        case (Some ts)
        hence t: "\<not>(cs = [] \<and> t = None)" and some_t: "\<not> t = None" by auto
        from True have empty: "\<not> cs \<noteq> []" by simp
        show ?thesis by (auto simp: if_not_P[OF t] if_not_P[OF empty] if_not_P[OF some_t] shows_attrs_assoc)
      qed
    next
      case False
      hence ne: "\<not>(cs = [] \<and> t = None)" by simp
      show ?thesis
      proof (cases t)
        case None show ?thesis by (auto simp: if_not_P[OF ne] if_P[OF False] if_P[OF None] shows_attrs_assoc IH)
      next
        case (Some ts)
        hence some_t: "\<not> t = None" by simp
        show ?thesis by (auto simp: if_not_P[OF ne] if_P[OF False] if_not_P[OF some_t] shows_attrs_assoc IH)
      qed
    qed
  qed

  lemma assoc_xml: "shows_prec d (xml::xml) r @ s = shows_prec d xml (r @ s)" unfolding shows_prec_xml_def by (rule assoc_xml')
  lemma assoc_xml_elt: "\<forall>(x::xml)\<in>set xs. shows_prec d x r @ s = shows_prec d x (r @ s)" using assoc_xml by auto

  instance by
    (intro_classes,rule assoc_xml,unfold shows_list_xml_def)
    (rule shows_list_aux_assoc[OF assoc_xml_elt])
end

instantiation xmldoc :: "show" begin
  fun shows_xmldoc where "shows_xmldoc(XMLDOC h x) = (h +#+ shows_nl +@+ shows x)"
  definition "shows_prec (d::nat) doc = shows_xmldoc doc"
  definition "shows_list (xs::xmldoc list) \<equiv> shows_list_aux shows xs"

  lemma assoc_xmldoc: "shows_prec d (x::xmldoc) r @ s = shows_prec d x (r @ s)" by (cases x) (auto simp: shows_prec_xmldoc_def)
  lemma assoc_xmldoc_elt: "\<forall>(x::xmldoc)\<in>set xs. shows_prec d x r @ s = shows_prec d x (r @ s)" using assoc_xmldoc by auto

  instance by
    (intro_classes,rule assoc_xmldoc,unfold shows_list_xmldoc_def)
    (rule shows_list_aux_assoc[OF assoc_xmldoc_elt])
end


subsection {* XML Transformers *}
hide_const (open) return
hide_const (open) error

(* transformers on xml nodes *)
type_synonym
  'a xmlt = "xml \<Rightarrow> string + 'a"

definition
  map :: "'a xmlt \<Rightarrow> xml list \<Rightarrow> string + 'a list"
where
  "map = Error_Monad.mapM"
hide_const (open) map

fun
  get_name :: "xml \<Rightarrow> string" 
where
  "get_name (XML name _ _ _) = name"
hide_const (open) get_name

definition
  string2bool :: "string \<Rightarrow> string + bool"
where
  "string2bool s \<equiv> (if s = ''true''
     then Error_Monad.return True
     else (if s = ''false''
       then Error_Monad.return False
       else Error_Monad.error (''cannot convert '' @ s @ '' into Boolean'')))"

fun
  "text" :: "string \<Rightarrow> string xmlt"
where
  "text tag (XML n atts cs t) = 
     (if n = tag \<and> atts = [] \<and> cs = [] \<and> t \<noteq> None \<and> the t \<noteq> []
       then Error_Monad.return (the t)
       else Error_Monad.error (concat [''could not extract text for '', tag,'' from '',
         [newline], to_string (XML n atts cs t)]))"
hide_const (open) "text"

fun
 bool :: "string \<Rightarrow> bool xmlt"
where
  "bool tag node = (Xml.text tag node \<guillemotright>= string2bool)"
hide_const (open) bool

definition
  no_transform :: "string \<Rightarrow> xml \<Rightarrow> string + 'a"
where
  "no_transform tag xml = error (concat [''could not transform the following
     xml element (expected '', tag, '')'', [newline], to_string xml])"

fun
  xml0elements :: "string \<Rightarrow> 'a \<Rightarrow> 'a xmlt"
where
  "xml0elements tag x (XML name atts cs text) = 
     (if name = tag \<and> atts = [] \<and> cs = [] \<and> text = None
       then return x 
       else no_transform tag (XML name atts cs text))"

fun
  xml1element :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b xmlt"
where
  "xml1element tag p1 f (XML name atts cs text) = 
     (if name = tag \<and> atts = [] \<and> length cs = 1 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         return (f x1)
       } else no_transform tag (XML name atts cs text))"

fun
  xml2elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml2elements tag p1 p2 f (XML name atts cs text) = 
     (if name = tag \<and> atts = [] \<and> length cs = 2 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         return (f x1 x2)
       } else no_transform tag (XML name atts cs text))"

fun
  xml3elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml3elements tag p1 p2 p3 f (XML name atts cs text) = 
     (if name = tag \<and> atts = [] \<and> length cs = 3 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         x3 \<leftarrow> p3 (cs ! 2);
         return (f x1 x2 x3)
       } else no_transform tag (XML name atts cs text))"

fun
  xml4elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> 'e xmlt"
where
  "xml4elements tag p1 p2 p3 p4 f (XML name atts cs text) = 
     (if name = tag \<and> atts = [] \<and> length cs = 4 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         x3 \<leftarrow> p3 (cs ! 2);
         x4 \<leftarrow> p4 (cs ! 3);
         return (f x1 x2 x3 x4)
       } else no_transform tag (XML name atts cs text))"

fun
  xml5elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> 'e xmlt
     \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e \<Rightarrow> 'f) \<Rightarrow> 'f xmlt"
where
  "xml5elements tag p1 p2 p3 p4 p5 f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> length cs = 5 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         x3 \<leftarrow> p3 (cs ! 2);
         x4 \<leftarrow> p4 (cs ! 3);
         x5 \<leftarrow> p5 (cs ! 4);
         return (f x1 x2 x3 x4 x5)
       } else no_transform tag (XML name atts cs text))"

fun
  xml0to1elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a option \<Rightarrow> 'b) \<Rightarrow> 'b xmlt"
where
  "xml0to1elements tag p1 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 0 \<and> l \<le> 1 \<and> text = None
       then do {
         (if l = 1
           then do {
             x1 \<leftarrow> p1 (cs ! 0);
             return (f (Some x1))
           } else return (f None))
       } else no_transform tag (XML name atts cs text)))"

fun
  xml1to2elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> ('a \<Rightarrow> 'b option \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml1to2elements tag p1 p2 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 1 \<and> l \<le> 2 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         (if l = 2
           then do {
             x2 \<leftarrow> p2 (cs ! 1);
             return (f x1 (Some x2))
           } else return (f x1 None))
       } else no_transform tag (XML name atts cs text)))"

(* parses the first element, the second argument must be a boolean. If the boolean is true, then the last element is parsed additionally *)
fun
  xml2nd_choice :: "string \<Rightarrow> 'a xmlt \<Rightarrow> string \<Rightarrow> 'b xmlt \<Rightarrow> ('a \<Rightarrow> 'b option \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml2nd_choice tag p1 cn p2 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 2 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         b \<leftarrow> Xml.bool cn (cs ! 1);
         (if b
           then do {
             x2 \<leftarrow> p2 (cs ! (l - 1));
             return (f x1 (Some x2))
           } else return (f x1 None))
       } else no_transform tag (XML name atts cs text)))"

fun
  xml2to3elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c option \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml2to3elements tag p1 p2 p3 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 2 \<and> l \<le> 3 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         (if l = 3
           then do {
             x3 \<leftarrow> p3 (cs ! 2);
             return (f x1 x2 (Some x3))
           } else return (f x1 x2 None))
       } else no_transform tag (XML name atts cs text)))"

fun
  xml3to4elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c option \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> 'e xmlt"
where
  "xml3to4elements tag p1 p2 p3 p4 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 3 \<and> l \<le> 4 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         (if l = 4
           then do {
             x3 \<leftarrow> p3 (cs ! 2);
             x4 \<leftarrow> p4 (cs ! 3);
             return (f x1 x2 (Some x3) x4)
           } else do {
             x4 \<leftarrow> p4 (cs ! 2);
             return (f x1 x2 None x4)
           } )
       } else no_transform tag (XML name atts cs text)))"

fun
  xml_many_elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a list \<Rightarrow> 'b) \<Rightarrow> 'b xmlt"
where
  "xml_many_elements tag p f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> text = None
       then (Xml.map p cs \<guillemotright>= (return \<circ> f))
       else no_transform tag (XML name atts cs text))"

fun
  xml1many_elements_gen :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a \<Rightarrow> 'b xmlt) \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml1many_elements_gen tag p1 p2 f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> text = None \<and> cs \<noteq> []
       then do {
         x \<leftarrow> p1 (hd cs);
         xs \<leftarrow> Xml.map (p2 x) (tl cs);
         return (f x xs)
       } else no_transform tag (XML name atts cs text))"

definition
  xml1many_elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml1many_elements tag p1 p2 = xml1many_elements_gen tag p1 (\<lambda>_. p2)"

fun
  xml2many_elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c list \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml2many_elements tag p1 p2 p3 f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> text = None \<and> length cs \<ge> 2
       then do {
         x \<leftarrow> p1 (cs ! 0);
         y \<leftarrow> p2 (cs ! 1);
         xs \<leftarrow> Xml.map p3 (tl (tl cs));
         return (f x y xs)
       } else no_transform tag (XML name atts cs text))"

fun
  xml1or2many_elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a \<Rightarrow> 'b option \<Rightarrow> 'c list \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml1or2many_elements tag p1 p2 p3 f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> text = None \<and> length cs \<ge> 2
       then do {
         x \<leftarrow> p1 (cs ! 0);
         try do {
           y \<leftarrow> p2 (cs ! 1);
           xs \<leftarrow> Xml.map p3 (tl (tl cs));
           return (f x (Some y) xs)
         }  catch (\<lambda> _. 
               do {
                 xs \<leftarrow> Xml.map p3 (tl cs);
                 return (f x None xs)
                  })
       } else no_transform tag (XML name atts cs text))"


fun
  xml1many2elements_gen :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a \<Rightarrow> 'b xmlt) \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> 'e xmlt"
where
  "xml1many2elements_gen tag p1 p2 p3 p4 f (XML name atts cs text) = (
     let ds = List.rev cs; l = length cs in
     (if name = tag \<and> atts = [] \<and> text = None \<and> l \<ge> 3
       then do {
         x \<leftarrow> p1 (cs ! 0);
         xs \<leftarrow> Xml.map (p2 x) (tl (take (l - 2) cs));
         y \<leftarrow> p3 (ds ! 1);
         z \<leftarrow> p4 (ds ! 0);
         return (f x xs y z)
       } else no_transform tag (XML name atts cs text)))"

fun
  xml1many2elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> 'e xmlt"
where
  "xml1many2elements tag p1 p2 = xml1many2elements_gen tag p1 (\<lambda>_. p2)"

fun
  xml_many2elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a list \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml_many2elements tag p1 p2 p3 f (XML name atts cs text) = (
     let ds = List.rev cs in
     (if name = tag \<and> atts = [] \<and> text = None \<and> length cs \<ge> 2
       then do {
         xs \<leftarrow> Xml.map p1 (List.rev (tl (tl ds)));
         y \<leftarrow> p2 (ds ! 1);
         z \<leftarrow> p3 (ds ! 0);
         return (f xs y z)
       } else no_transform tag (XML name atts cs text)))"

fun
  xml_options :: "(string \<times> 'a xmlt) list \<Rightarrow> 'a xmlt"
where
  "xml_options ps x =
     (case lookup (Xml.get_name x) ps of
       None \<Rightarrow> error (concat [''expected one of: '', concat (map (\<lambda>p. fst p @ '' '') ps), [newline], ''but found'', [newline], to_string x])
     | Some p \<Rightarrow> p x)"

fun
  xml_choice :: "string \<Rightarrow> 'a xmlt list \<Rightarrow> 'a xmlt"
where
  "xml_choice e [] x = error (concat [''error in parsing choice for '',e,[newline],to_string x])"
| "xml_choice e (p # ps) x = (case p x of Inr r \<Rightarrow> Inr r | _ \<Rightarrow> xml_choice e ps x)"

fun
  xml_change :: "'a xmlt \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b xmlt"
where
  "xml_change p f x = (p x \<guillemotright>= (\<lambda>a. return (f a)))"

fun
  digit2int :: "char \<Rightarrow> string + int"
where
  "digit2int x =
     (if x = CHR ''0'' then return 0 else
     (if x = CHR ''1'' then return 1 else
     (if x = CHR ''2'' then return 2 else
     (if x = CHR ''3'' then return 3 else
     (if x = CHR ''4'' then return 4 else
     (if x = CHR ''5'' then return 5 else
     (if x = CHR ''6'' then return 6 else
     (if x = CHR ''7'' then return 7 else
     (if x = CHR ''8'' then return 8 else
     (if x = CHR ''9'' then return 9 else 
       error ([x] @ '' is not a digit'')))))))))))"

fun
  string2int_aux :: "int \<Rightarrow> string \<Rightarrow> string + int" 
where
  "string2int_aux n [] = return n" |
  "string2int_aux n (d # s) = (digit2int d \<guillemotright>= (\<lambda>m. string2int_aux (10 * n + m) s))"

definition
  string2int :: "string \<Rightarrow> string + int"
where
  "string2int s =
     (if s = []
       then error ''cannot convert empty string into number'' 
       else if take 1 s = ''-''
         then string2int_aux 0 (tl s) \<guillemotright>= (\<lambda> i. return (0 - i))
         else string2int_aux 0 s)"

hide_const string2int_aux

fun
  int :: "string \<Rightarrow> int xmlt"
where
  "int tag x = (Xml.text tag x \<guillemotright>= string2int)"
hide_const (open) int

fun
  nat :: "string \<Rightarrow> nat xmlt"
where
  "nat tag x = do {
    txt \<leftarrow> Xml.text tag x;
    i \<leftarrow> string2int txt;
    return (Int.nat i)
  }"
hide_const (open) nat

definition
  rat :: "rat xmlt"
where
  "rat \<equiv> xml_options [(''integer'', xml_change (Xml.int ''integer'') (\<lambda>x. Fract x 1)),
        (''rational'', xml2elements ''rational'' (Xml.int ''numerator'') (Xml.int ''denominator'') Fract)]"
hide_const (open) rat
end